import ibis.common.exceptions as com
from ibis.config import options

# these objects are exposed in the public API and are not used in the module
from ibis.impala.client import (  # noqa: F401
    ImpalaClient,
    ImpalaConnection,
    ImpalaDatabase,
    ImpalaTable,
)
from ibis.impala.compiler import dialect  # noqa: F401
from ibis.impala.udf import *  # noqa: F401,F403


def compile(expr, params=None):
    """Force compilation of expression.

    Returns
    -------
    str

    """
    from ibis.impala.compiler import to_sql

    return to_sql(expr, dialect.make_context(params=params))


def verify(expr, params=None):
    """
    Determine if expression can be successfully translated to execute on Impala
    """
    try:
        compile(expr, params=params)
        return True
    except com.TranslationError:
        return False


def connect(
    host='localhost',
    port=21050,
    database='default',
    timeout=45,
    use_ssl=False,
    ca_cert=None,
    user=None,
    password=None,
    auth_mechanism='NOSASL',
    kerberos_service_name='impala',
    pool_size=8,
    hdfs_client=None,
):
    """Create an ImpalaClient for use with Ibis.

    Parameters
    ----------
    host : str, optional
        Host name of the impalad or HiveServer2 in Hive
    port : int, optional
        Impala's HiveServer2 port
    database : str, optional
        Default database when obtaining new cursors
    timeout : int, optional
        Connection timeout in seconds when communicating with HiveServer2
    use_ssl : bool, optional
        Use SSL when connecting to HiveServer2
    ca_cert : str, optional
        Local path to 3rd party CA certificate or copy of server certificate
        for self-signed certificates. If SSL is enabled, but this argument is
        ``None``, then certificate validation is skipped.
    user : str, optional
        LDAP user to authenticate
    password : str, optional
        LDAP password to authenticate
    auth_mechanism : str, optional
        {'NOSASL' <- default, 'PLAIN', 'GSSAPI', 'LDAP'}.
        Use NOSASL for non-secured Impala connections.  Use PLAIN for
        non-secured Hive clusters.  Use LDAP for LDAP authenticated
        connections.  Use GSSAPI for Kerberos-secured clusters.
    kerberos_service_name : str, optional
        Specify particular impalad service principal.

    Examples
    --------
    >>> import ibis
    >>> import os
    >>> hdfs_host = os.environ.get('IBIS_TEST_NN_HOST', 'localhost')
    >>> hdfs_port = int(os.environ.get('IBIS_TEST_NN_PORT', 50070))
    >>> impala_host = os.environ.get('IBIS_TEST_IMPALA_HOST', 'localhost')
    >>> impala_port = int(os.environ.get('IBIS_TEST_IMPALA_PORT', 21050))
    >>> hdfs = ibis.hdfs_connect(host=hdfs_host, port=hdfs_port)
    >>> hdfs  # doctest: +ELLIPSIS
    <ibis.filesystems.WebHDFS object at 0x...>
    >>> client = ibis.impala.connect(
    ...     host=impala_host,
    ...     port=impala_port,
    ...     hdfs_client=hdfs,
    ... )
    >>> client  # doctest: +ELLIPSIS
    <ibis.impala.client.ImpalaClient object at 0x...>

    Returns
    -------
    ImpalaClient
    """
    params = {
        'host': host,
        'port': port,
        'database': database,
        'timeout': timeout,
        'use_ssl': use_ssl,
        'ca_cert': ca_cert,
        'user': user,
        'password': password,
        'auth_mechanism': auth_mechanism,
        'kerberos_service_name': kerberos_service_name,
    }

    con = ImpalaConnection(pool_size=pool_size, **params)
    try:
        client = ImpalaClient(con, hdfs_client=hdfs_client)
    except Exception:
        con.close()
        raise
    else:
        if options.default_backend is None:
            options.default_backend = client

    return client



from typing import Iterable, TypeVar

import ibis.expr.types as ir
from ibis.common import exceptions as ex

IS_SHAPELY_AVAILABLE = False
try:
    import shapely

    IS_SHAPELY_AVAILABLE = True
except ImportError:
    ...

NumberType = TypeVar('NumberType', int, float)
# Geometry primitives (2D)
PointType = Iterable[NumberType]
LineStringType = Iterable[PointType]
PolygonType = Iterable[LineStringType]
# Multipart geometries (2D)
MultiPointType = Iterable[PointType]
MultiLineStringType = Iterable[LineStringType]
MultiPolygonType = Iterable[PolygonType]


def _format_point_value(value: PointType) -> str:
    """Convert a iterable with a point to text."""
    return ' '.join(str(v) for v in value)


def _format_linestring_value(value: LineStringType, nested=False) -> str:
    """Convert a iterable with a linestring to text."""
    template = '({})' if nested else '{}'
    if not isinstance(value[0], (tuple, list)):
        msg = '{} structure expected: LineStringType'.format(
            'Data' if not nested else 'Inner data'
        )
        raise ex.IbisInputError(msg)
    return template.format(
        ', '.join(_format_point_value(point) for point in value)
    )


def _format_polygon_value(value: PolygonType, nested=False) -> str:
    """Convert a iterable with a polygon to text."""
    template = '({})' if nested else '{}'
    if not isinstance(value[0][0], (tuple, list)):
        msg = '{} data structure expected: PolygonType'.format(
            'Data' if not nested else 'Inner data'
        )
        raise ex.IbisInputError(msg)

    return template.format(
        ', '.join(
            _format_linestring_value(line, nested=True) for line in value
        )
    )


def _format_multipoint_value(value: MultiPointType) -> str:
    """Convert a iterable with a multipoint to text."""
    if not isinstance(value[0], (tuple, list)):
        raise ex.IbisInputError('Data structure expected: MultiPointType')
    return ', '.join(
        '({})'.format(_format_point_value(point)) for point in value
    )


def _format_multilinestring_value(value: MultiLineStringType) -> str:
    """Convert a iterable with a multilinestring to text."""
    if not isinstance(value[0][0], (tuple, list)):
        raise ex.IbisInputError('Data structure expected: MultiLineStringType')
    return ', '.join(
        '({})'.format(_format_linestring_value(line)) for line in value
    )


def _format_multipolygon_value(value: MultiPolygonType) -> str:
    """Convert a iterable with a multipolygon to text."""
    if not isinstance(value[0][0], (tuple, list)):
        raise ex.IbisInputError('Data structure expected: MultiPolygonType')
    return ', '.join(
        _format_polygon_value(polygon, nested=True) for polygon in value
    )


def _format_geo_metadata(op, value: str, inline_metadata: bool = False) -> str:
    """Format a geometry/geography text when it is necessary."""
    srid = op.args[1].srid
    geotype = op.args[1].geotype

    if inline_metadata:
        value = "'{}{}'{}".format(
            'SRID={};'.format(srid) if srid else '',
            value,
            '::{}'.format(geotype) if geotype else '',
        )
        return value

    geofunc = (
        'ST_GeogFromText' if geotype == 'geography' else 'ST_GeomFromText'
    )

    value = repr(value)
    if srid:
        value += ', {}'.format(srid)

    return "{}({})".format(geofunc, value)


def translate_point(value: Iterable) -> str:
    """Translate a point to WKT."""
    return "POINT ({})".format(_format_point_value(value))


def translate_linestring(value: Iterable) -> str:
    """Translate a linestring to WKT."""
    return "LINESTRING ({})".format(_format_linestring_value(value))


def translate_polygon(value: Iterable) -> str:
    """Translate a polygon to WKT."""
    return "POLYGON ({})".format(_format_polygon_value(value))


def translate_multilinestring(value: Iterable) -> str:
    """Translate a multilinestring to WKT."""
    return "MULTILINESTRING ({})".format(_format_multilinestring_value(value))


def translate_multipoint(value: Iterable) -> str:
    """Translate a multipoint to WKT."""
    return "MULTIPOINT ({})".format(_format_multipoint_value(value))


def translate_multipolygon(value: Iterable) -> str:
    """Translate a multipolygon to WKT."""
    return "MULTIPOLYGON ({})".format(_format_multipolygon_value(value))


def translate_literal(expr, inline_metadata: bool = False) -> str:
    op = expr.op()
    value = op.value

    if IS_SHAPELY_AVAILABLE and isinstance(
        value, shapely.geometry.base.BaseGeometry
    ):
        result = value.wkt
    elif isinstance(expr, ir.PointScalar):
        result = translate_point(value)
    elif isinstance(expr, ir.LineStringScalar):
        result = translate_linestring(value)
    elif isinstance(expr, ir.PolygonScalar):
        result = translate_polygon(value)
    elif isinstance(expr, ir.MultiLineStringScalar):
        result = translate_multilinestring(value)
    elif isinstance(expr, ir.MultiPointScalar):
        result = translate_multipoint(value)
    elif isinstance(expr, ir.MultiPolygonScalar):
        result = translate_multipolygon(value)
    else:
        raise ex.UnboundExpressionError('Geo Spatial type not supported.')
    return _format_geo_metadata(op, result, inline_metadata)



""" Tests for geo spatial data types"""
import pytest

import ibis
from ibis.expr.tests.mocks import (
    GeoMockConnectionOmniSciDB,
    GeoMockConnectionPostGIS,
)

pytest.importorskip('geoalchemy2')
pytest.importorskip('shapely')
pytest.importorskip('geopandas')

pytest.mark.postgis
pytest.mark.omniscidb

mock_omniscidb = GeoMockConnectionOmniSciDB()
mock_postgis = GeoMockConnectionPostGIS()


@pytest.mark.parametrize('backend', [mock_postgis])
@pytest.mark.parametrize(
    'modifier',
    [
        {},
        {'srid': '4326'},
        {'srid': '4326', 'geo_type': 'geometry'},
        {'srid': '4326', 'geo_type': 'geography'},
    ],
)
@pytest.mark.parametrize(
    'shape,value,expected',
    [
        # Geometry primitives (2D)
        ('point', (30, 10), 'POINT (30 10)'),
        (
            'linestring',
            ((30, 10), (10, 30), (40, 40)),
            'LINESTRING (30 10, 10 30, 40 40)',
        ),
        (
            'polygon',
            (
                ((35, 10), (45, 45), (15, 40), (10, 20), (35, 10)),
                ((20, 30), (35, 35), (30, 20), (20, 30)),
            ),
            (
                'POLYGON ((35 10, 45 45, 15 40, 10 20, 35 10), '
                + '(20 30, 35 35, 30 20, 20 30))'
            ),
        ),
        (
            'polygon',
            (((30, 10), (40, 40), (20, 40), (10, 20), (30, 10)),),
            'POLYGON ((30 10, 40 40, 20 40, 10 20, 30 10))',
        ),
        # Multipart geometries (2D)
        (
            'multipoint',
            ((10, 40), (40, 30), (20, 20), (30, 10)),
            'MULTIPOINT ((10 40), (40 30), (20 20), (30 10))',
        ),
        (
            'multilinestring',
            (
                ((10, 10), (20, 20), (10, 40)),
                ((40, 40), (30, 30), (40, 20), (30, 10)),
            ),
            (
                'MULTILINESTRING ((10 10, 20 20, 10 40), '
                + '(40 40, 30 30, 40 20, 30 10))'
            ),
        ),
        (
            'multipolygon',
            (
                (((40, 40), (20, 45), (45, 30), (40, 40)),),
                (
                    (
                        (20, 35),
                        (10, 30),
                        (10, 10),
                        (30, 5),
                        (45, 20),
                        (20, 35),
                    ),
                    ((30, 20), (20, 15), (20, 25), (30, 20)),
                ),
            ),
            (
                'MULTIPOLYGON (((40 40, 20 45, 45 30, 40 40)), '
                + '((20 35, 10 30, 10 10, 30 5, 45 20, 20 35), '
                + '(30 20, 20 15, 20 25, 30 20)))'
            ),
        ),
    ],
)
def test_geo_literals_smoke(backend, shape, value, modifier, expected):
    """Smoke tests for geo spatial literals"""
    expr_type = '{}{}{}'.format(
        shape,
        ';{}'.format(modifier['srid']) if 'srid' in modifier else '',
        ':{}'.format(modifier['geo_type']) if 'geo_type' in modifier else '',
    )
    expr = ibis.literal(value, type=expr_type).name('tmp')
    result_expected = "SELECT '{}{}'{}".format(
        'SRID={};'.format(modifier['srid']) if 'srid' in modifier else '',
        expected,
        '::{}'.format(modifier['geo_type']) if 'geo_type' in modifier else '',
    )

    assert str(backend.compile(expr)) == result_expected


@pytest.mark.parametrize(
    'fn_expr',
    [
        pytest.param(lambda t: t.geo_point.srid(), id='point_srid'),
        pytest.param(
            lambda t: t.geo_point.set_srid(4326), id='point_set_srid'
        ),
        pytest.param(lambda t: t.geo_point.x(), id='point_x'),
        pytest.param(lambda t: t.geo_point.y(), id='point_y'),
        pytest.param(
            lambda t: t.geo_linestring.contains(t.geo_point),
            id='linestring_contains',
        ),
        pytest.param(
            lambda t: t.geo_linestring.end_point(), id='linestring_end_point'
        ),
        pytest.param(
            lambda t: t.geo_linestring.length(), id='linestring_length'
        ),
        pytest.param(
            lambda t: t.geo_linestring.max_distance(t.geo_point),
            id='linestring_max_distance',
        ),
        pytest.param(
            lambda t: t.geo_linestring.point_n(1), id='linestring_point_n'
        ),
        pytest.param(
            lambda t: t.geo_linestring.start_point(),
            id='linestring_start_point',
        ),
        pytest.param(
            lambda t: t.geo_linestring.x_max(), id='linestring_x_max'
        ),
        pytest.param(
            lambda t: t.geo_linestring.x_min(), id='linestring_x_min'
        ),
        pytest.param(
            lambda t: t.geo_linestring.y_max(), id='linestring_y_max'
        ),
        pytest.param(
            lambda t: t.geo_linestring.y_min(), id='linestring_y_min'
        ),
        pytest.param(lambda t: t.geo_polygon.area(), id='polygon_area'),
        pytest.param(
            lambda t: t.geo_polygon.perimeter(), id='polygon_perimeter'
        ),
        pytest.param(
            lambda t: t.geo_multipolygon.n_points(), id='multipolygon_n_points'
        ),
        pytest.param(
            lambda t: t.geo_multipolygon.n_rings(), id='multipolygon_n_rings'
        ),
        # TODO: the mock tests don't support multipoint and multilinestring
        #       yet, but once they do, add some more tests here.
    ],
)
@pytest.mark.parametrize('backend', [mock_omniscidb, mock_postgis])
@pytest.mark.xfail_unsupported
def test_geo_ops_smoke(backend, fn_expr):
    """Smoke tests for geo spatial operations."""
    geo_table = backend.table('geo')
    assert fn_expr(geo_table).compile() != ''



import pandas.util.testing as tm
import pytest
from numpy import testing

gp = pytest.importorskip('geopandas')
sa = pytest.importorskip('sqlalchemy')
pytest.importorskip('psycopg2')

pytestmark = [pytest.mark.postgis, pytest.mark.postgres_extensions]


def test_load_geodata(con):
    t = con.table('geo')
    result = t.execute()
    assert isinstance(result, gp.GeoDataFrame)


def test_empty_select(geotable):
    expr = geotable[geotable.geo_point.equals(geotable.geo_linestring)]
    result = expr.execute()
    assert len(result) == 0


def test_select_point_geodata(geotable):
    expr = geotable['geo_point']
    sqla_expr = expr.compile()
    compiled = str(sqla_expr.compile(compile_kwargs=dict(literal_binds=True)))
    expected = "SELECT ST_AsEWKB(t0.geo_point) AS geo_point \nFROM geo AS t0"
    assert compiled == expected
    data = expr.execute()
    assert data.geom_type.iloc[0] == 'Point'


def test_select_linestring_geodata(geotable):
    expr = geotable['geo_linestring']
    sqla_expr = expr.compile()
    compiled = str(sqla_expr.compile(compile_kwargs=dict(literal_binds=True)))
    expected = (
        "SELECT ST_AsEWKB(t0.geo_linestring) AS geo_linestring \n"
        "FROM geo AS t0"
    )
    assert compiled == expected
    data = expr.execute()
    assert data.geom_type.iloc[0] == 'LineString'


def test_select_polygon_geodata(geotable):
    expr = geotable['geo_polygon']
    sqla_expr = expr.compile()
    compiled = str(sqla_expr.compile(compile_kwargs=dict(literal_binds=True)))
    expected = (
        "SELECT ST_AsEWKB(t0.geo_polygon) AS geo_polygon \n" "FROM geo AS t0"
    )
    assert compiled == expected
    data = expr.execute()
    assert data.geom_type.iloc[0] == 'Polygon'


def test_select_multipolygon_geodata(geotable):
    expr = geotable['geo_multipolygon']
    sqla_expr = expr.compile()
    compiled = str(sqla_expr.compile(compile_kwargs=dict(literal_binds=True)))
    expected = (
        "SELECT ST_AsEWKB(t0.geo_multipolygon) AS geo_multipolygon \n"
        "FROM geo AS t0"
    )
    assert compiled == expected
    data = expr.execute()
    assert data.geom_type.iloc[0] == 'MultiPolygon'


def test_geo_area(geotable, gdf):
    expr = geotable.geo_multipolygon.area()
    result = expr.execute()
    expected = gp.GeoSeries(gdf.geo_multipolygon).area
    tm.assert_series_equal(result, expected, check_names=False)


def test_geo_buffer(geotable, gdf):
    expr = geotable.geo_linestring.buffer(1.0)
    result = expr.execute()
    expected = gp.GeoSeries(gdf.geo_linestring).buffer(1.0)
    tm.assert_series_equal(
        result.area, expected.area, check_names=False, check_less_precise=2
    )


def test_geo_contains(geotable):
    expr = geotable.geo_point.buffer(1.0).contains(geotable.geo_point)
    assert expr.execute().all()


def test_geo_contains_properly(geotable):
    expr = geotable.geo_point.buffer(1.0).contains_properly(geotable.geo_point)
    assert expr.execute().all()


def test_geo_covers(geotable):
    expr = geotable.geo_point.buffer(1.0).covers(geotable.geo_point)
    assert expr.execute().all()


def test_geo_covered_by(geotable):
    expr = geotable.geo_point.covered_by(geotable.geo_point.buffer(1.0))
    assert expr.execute().all()


def test_geo_d_fully_within(geotable):
    expr = geotable.geo_point.d_fully_within(
        geotable.geo_point.buffer(1.0), 2.0
    )
    assert expr.execute().all()


def test_geo_d_within(geotable):
    expr = geotable.geo_point.d_within(geotable.geo_point.buffer(1.0), 1.0)
    assert expr.execute().all()


def test_geo_envelope(geotable, gdf):
    expr = geotable.geo_linestring.buffer(1.0).envelope()
    result = expr.execute()
    expected = gp.GeoSeries(gdf.geo_linestring).buffer(1.0).envelope
    tm.assert_series_equal(result.area, expected.area, check_names=False)


def test_geo_within(geotable):
    expr = geotable.geo_point.within(geotable.geo_point.buffer(1.0))
    assert expr.execute().all()


def test_geo_disjoint(geotable):
    expr = geotable.geo_point.disjoint(geotable.geo_point)
    assert not expr.execute().any()


def test_geo_equals(geotable):
    expr = geotable.geo_point.equals(geotable.geo_point)
    assert expr.execute().all()


def test_geo_geometry_n(geotable, gdf):
    expr = geotable.geo_multipolygon.geometry_n(1)  # PostGIS is one-indexed.
    result = expr.execute()
    first_polygon = gdf.apply(lambda x: list(x.geo_multipolygon)[0], axis=1)
    for a, b in zip(result, first_polygon):
        assert a.equals(b)


def test_geo_geometry_type(geotable):
    expr = geotable.geo_point.geometry_type()
    assert (expr.execute() == 'ST_Point').all()
    expr = geotable.geo_multipolygon.geometry_type()
    assert (expr.execute() == 'ST_MultiPolygon').all()


def test_geo_intersects(geotable):
    expr = geotable.geo_point.intersects(geotable.geo_point.buffer(1.0))
    assert expr.execute().all()


def test_geo_is_valid(geotable):
    expr = geotable.geo_point.is_valid()
    assert expr.execute().all()


def test_geo_line_locate_point(geotable):
    expr = geotable.geo_linestring.line_locate_point(geotable.geo_point)
    assert (expr.execute() == 0).all()


def test_geo_line_merge(geotable, gdf):
    expr = geotable.geo_linestring.line_merge()
    expected = gp.GeoSeries(gdf.geo_linestring)
    tm.assert_series_equal(expr.execute().length, expected.length)


def test_geo_line_substring(geotable, gdf):
    expr = geotable.geo_linestring.line_substring(0.25, 0.75)
    result = expr.execute()
    expected = gp.GeoSeries(gdf.geo_linestring)
    tm.assert_series_equal(expected.length / 2.0, result.length)


def test_geo_ordering_equals(geotable):
    expr = geotable.geo_point.ordering_equals(geotable.geo_point)
    assert expr.execute().all()


def test_geo_overlaps(geotable):
    expr = geotable.geo_point.overlaps(geotable.geo_point.buffer(1.0))
    assert not expr.execute().any()


def test_geo_touches(geotable):
    expr = geotable.geo_point.touches(geotable.geo_linestring)
    assert expr.execute().all()


def test_geo_distance(geotable, gdf):
    expr = geotable.geo_point.distance(geotable.geo_multipolygon.centroid())
    result = expr.execute()
    expected = gdf.geo_point.distance(
        gp.GeoSeries(gdf.geo_multipolygon).centroid
    )
    tm.assert_series_equal(result, expected, check_names=False)


def test_geo_length(geotable, gdf):
    expr = geotable.geo_linestring.length()
    result = expr.execute()
    expected = gp.GeoSeries(gdf.geo_linestring).length
    tm.assert_series_equal(result, expected, check_names=False)


def test_geo_n_points(geotable):
    expr = geotable.geo_linestring.n_points()
    result = expr.execute()
    assert (result == 2).all()


def test_geo_perimeter(geotable):
    expr = geotable.geo_multipolygon.perimeter()
    result = expr.execute()
    # Geopandas doesn't implement perimeter, so we do a simpler check.
    assert (result > 0.0).all()


def test_geo_srid(geotable):
    expr = geotable.geo_linestring.srid()
    result = expr.execute()
    assert (result == 0).all()


def test_geo_difference(geotable, gdf):
    expr = (
        geotable.geo_linestring.buffer(1.0)
        .difference(geotable.geo_point.buffer(0.5))
        .area()
    )
    result = expr.execute()
    expected = (
        gp.GeoSeries(gdf.geo_linestring)
        .buffer(1.0)
        .difference(gp.GeoSeries(gdf.geo_point).buffer(0.5))
        .area
    )
    tm.assert_series_equal(
        result, expected, check_names=False, check_less_precise=2
    )


def test_geo_intersection(geotable, gdf):
    expr = (
        geotable.geo_linestring.buffer(1.0)
        .intersection(geotable.geo_point.buffer(0.5))
        .area()
    )
    result = expr.execute()
    expected = (
        gp.GeoSeries(gdf.geo_linestring)
        .buffer(1.0)
        .intersection(gp.GeoSeries(gdf.geo_point).buffer(0.5))
        .area
    )
    tm.assert_series_equal(
        result, expected, check_names=False, check_less_precise=2
    )


def test_geo_unary_union(geotable, gdf):
    expr = geotable.geo_polygon.unary_union().area()
    expected = gp.GeoSeries(gdf.geo_polygon).unary_union.area
    testing.assert_almost_equal(expr.execute(), expected, decimal=2)


def test_geo_union(geotable, gdf):
    expr = geotable.geo_polygon.union(geotable.geo_multipolygon).area()
    expected = (
        gp.GeoSeries(gdf.geo_polygon)
        .union(gp.GeoSeries(gdf.geo_multipolygon))
        .area
    )
    tm.assert_series_equal(
        expr.execute(), expected, check_names=False, check_less_precise=2
    )


def test_geo_x(geotable, gdf):
    expr = geotable.geo_point.x()
    result = expr.execute()
    expected = gp.GeoSeries(gdf.geo_point).x
    tm.assert_series_equal(result, expected, check_names=False)


def test_geo_y(geotable, gdf):
    expr = geotable.geo_point.y()
    result = expr.execute()
    expected = gp.GeoSeries(gdf.geo_point).y
    tm.assert_series_equal(result, expected, check_names=False)



import pandas as pd
import pytest
from pytest import param

from ibis.tests.backends import Csv, Pandas, PySpark

# add here backends that passes join tests
all_db_join_supported = [Pandas, PySpark]


@pytest.mark.parametrize(
    'how',
    [
        'inner',
        'left',
        'right',
        'outer',
        param(
            'semi',
            marks=pytest.mark.xfail(
                raises=NotImplementedError, reason='Semi join not implemented'
            ),
        ),
        param(
            'anti',
            marks=pytest.mark.xfail(
                raises=NotImplementedError, reason='Anti join not implemented'
            ),
        ),
    ],
)
@pytest.mark.only_on_backends(all_db_join_supported)
# Csv is a subclass of Pandas so need to skip it explicitly.
@pytest.mark.skip_backends([Csv])
@pytest.mark.xfail_unsupported
def test_join_project_left_table(backend, con, batting, awards_players, how):

    left = batting[batting.yearID == 2015]
    right = awards_players[awards_players.lgID == 'NL'].drop(
        ['yearID', 'lgID']
    )

    left_df = left.execute()
    right_df = right.execute()
    predicate = ['playerID']
    result_order = ['playerID', 'yearID', 'lgID', 'stint']
    expr = left.join(right, predicate, how=how)[left]
    result = expr.execute().sort_values(result_order)

    joined = pd.merge(
        left_df, right_df, how=how, on=predicate, suffixes=('', '_y')
    ).sort_values(result_order)
    expected = joined[list(left.columns)]

    backend.assert_frame_equal(
        result[expected.columns], expected, check_like=True
    )



import pandas as pd

import ibis.expr.operations as ops
import ibis.expr.schema as sch
from ibis.file.client import FileClient
from ibis.pandas.core import execute, execute_node


def connect(path):
    """Create a HDF5Client for use with Ibis

    Parameters
    ----------
    path: str or pathlib.Path

    Returns
    -------
    HDF5Client
    """
    return HDFClient(path)


class HDFTable(ops.DatabaseTable):
    pass


class HDFClient(FileClient):
    extension = 'h5'
    table_class = HDFTable

    def insert(
        self, path, key, expr, format='table', data_columns=True, **kwargs
    ):

        path = self.root / path
        data = execute(expr)
        data.to_hdf(
            str(path), key, format=format, data_columns=data_columns, **kwargs
        )

    def table(self, name, path):
        if name not in self.list_tables(path):
            raise AttributeError(name)

        # get the schema
        with pd.HDFStore(str(path), mode='r') as store:
            df = store.select(name, start=0, stop=0)
            schema = sch.infer(df)

        t = self.table_class(name, schema, self).to_expr()
        self.dictionary[name] = path
        return t

    def list_tables(self, path=None):
        # tables are individual tables within a file

        if path is None:
            path = self.root

        if path.is_file() and str(path).endswith(self.extension):

            with pd.HDFStore(str(path), mode='r') as store:
                # strip leading /
                return [k[1:] for k in store.keys()]

        return []

    def list_databases(self, path=None):
        return self._list_databases_dirs_or_files(path)


@execute_node.register(HDFClient.table_class, HDFClient)
def hdf_read_table(op, client, scope, **kwargs):
    key = op.name
    path = client.dictionary[key]
    df = pd.read_hdf(str(path), key, mode='r')
    return df



import operator

import pandas as pd
import pandas.util.testing as tm
import pytest
from pytest import param

import ibis
from ibis.common.exceptions import IbisTypeError


def test_array_length(t, df):
    expr = t.projection(
        [
            t.array_of_float64.length().name('array_of_float64_length'),
            t.array_of_int64.length().name('array_of_int64_length'),
            t.array_of_strings.length().name('array_of_strings_length'),
        ]
    )
    result = expr.execute()
    expected = pd.DataFrame(
        {
            'array_of_float64_length': [2, 1, 0],
            'array_of_int64_length': [2, 0, 1],
            'array_of_strings_length': [2, 0, 1],
        }
    )

    tm.assert_frame_equal(result, expected)


def test_array_length_scalar(client):
    raw_value = [1, 2, 4]
    value = ibis.literal(raw_value)
    expr = value.length()
    result = client.execute(expr)
    expected = len(raw_value)
    assert result == expected


def test_array_collect(t, df):
    expr = t.group_by(t.dup_strings).aggregate(
        collected=t.float64_with_zeros.collect()
    )
    result = expr.execute().sort_values('dup_strings').reset_index(drop=True)
    expected = (
        df.groupby('dup_strings')
        .float64_with_zeros.apply(list)
        .reset_index()
        .rename(columns={'float64_with_zeros': 'collected'})
    )
    tm.assert_frame_equal(result, expected)


def test_array_collect_rolling_partitioned(t, df):
    window = ibis.trailing_window(1, order_by=t.plain_int64)
    colexpr = t.plain_float64.collect().over(window)
    expr = t['dup_strings', 'plain_int64', colexpr.name('collected')]
    result = expr.execute()
    expected = pd.DataFrame(
        {
            'dup_strings': ['d', 'a', 'd'],
            'plain_int64': [1, 2, 3],
            'collected': [[4.0], [4.0, 5.0], [5.0, 6.0]],
        }
    )[expr.columns]
    tm.assert_frame_equal(result, expected)


@pytest.mark.xfail(raises=IbisTypeError, reason='Not sure if this should work')
def test_array_collect_scalar(client):
    raw_value = 'abcd'
    value = ibis.literal(raw_value)
    expr = value.collect()
    result = client.execute(expr)
    expected = [raw_value]
    assert result == expected


@pytest.mark.parametrize(
    ['start', 'stop'],
    [
        (1, 3),
        (1, 1),
        (2, 3),
        (2, 5),
        (None, 3),
        (None, None),
        (3, None),
        # negative slices are not supported
        param(
            -3,
            None,
            marks=pytest.mark.xfail(
                raises=ValueError, reason='Negative slicing not supported'
            ),
        ),
        param(
            None,
            -3,
            marks=pytest.mark.xfail(
                raises=ValueError, reason='Negative slicing not supported'
            ),
        ),
        param(
            -3,
            -1,
            marks=pytest.mark.xfail(
                raises=ValueError, reason='Negative slicing not supported'
            ),
        ),
    ],
)
def test_array_slice(t, df, start, stop):
    expr = t.array_of_strings[start:stop]
    result = expr.execute()
    slicer = operator.itemgetter(slice(start, stop))
    expected = df.array_of_strings.apply(slicer)
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ['start', 'stop'],
    [
        (1, 3),
        (1, 1),
        (2, 3),
        (2, 5),
        (None, 3),
        (None, None),
        (3, None),
        # negative slices are not supported
        param(
            -3,
            None,
            marks=pytest.mark.xfail(
                raises=ValueError, reason='Negative slicing not supported'
            ),
        ),
        param(
            None,
            -3,
            marks=pytest.mark.xfail(
                raises=ValueError, reason='Negative slicing not supported'
            ),
        ),
        param(
            -3,
            -1,
            marks=pytest.mark.xfail(
                raises=ValueError, reason='Negative slicing not supported'
            ),
        ),
    ],
)
def test_array_slice_scalar(client, start, stop):
    raw_value = [-11, 42, 10]
    value = ibis.literal(raw_value)
    expr = value[start:stop]
    result = client.execute(expr)
    expected = raw_value[start:stop]
    assert result == expected


@pytest.mark.parametrize('index', [1, 3, 4, 11, -11])
def test_array_index(t, df, index):
    expr = t[t.array_of_float64[index].name('indexed')]
    result = expr.execute()
    expected = pd.DataFrame(
        {
            'indexed': df.array_of_float64.apply(
                lambda x: x[index] if -len(x) <= index < len(x) else None
            )
        }
    )
    tm.assert_frame_equal(result, expected)


@pytest.mark.parametrize('index', [1, 3, 4, 11])
def test_array_index_scalar(client, index):
    raw_value = [-10, 1, 2, 42]
    value = ibis.literal(raw_value)
    expr = value[index]
    result = client.execute(expr)
    expected = raw_value[index] if index < len(raw_value) else None
    assert result == expected


@pytest.mark.parametrize('n', [1, 3, 4, 7, -2])  # negative returns empty list
@pytest.mark.parametrize('mul', [lambda x, n: x * n, lambda x, n: n * x])
def test_array_repeat(t, df, n, mul):
    expr = t.projection([mul(t.array_of_strings, n).name('repeated')])
    result = expr.execute()
    expected = pd.DataFrame({'repeated': df.array_of_strings * n})
    tm.assert_frame_equal(result, expected)


@pytest.mark.parametrize('n', [1, 3, 4, 7, -2])  # negative returns empty list
@pytest.mark.parametrize('mul', [lambda x, n: x * n, lambda x, n: n * x])
def test_array_repeat_scalar(client, n, mul):
    raw_array = [1, 2]
    array = ibis.literal(raw_array)
    expr = mul(array, n)
    result = client.execute(expr)
    expected = mul(raw_array, n)
    assert result == expected


@pytest.mark.parametrize('op', [lambda x, y: x + y, lambda x, y: y + x])
def test_array_concat(t, df, op):
    x = t.array_of_float64.cast('array<string>')
    y = t.array_of_strings
    expr = op(x, y)
    result = expr.execute()
    expected = op(
        df.array_of_float64.apply(lambda x: list(map(str, x))),
        df.array_of_strings,
    )
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize('op', [lambda x, y: x + y, lambda x, y: y + x])
def test_array_concat_scalar(client, op):
    raw_left = [1, 2, 3]
    raw_right = [3, 4]
    left = ibis.literal(raw_left)
    right = ibis.literal(raw_right)
    expr = op(left, right)
    result = client.execute(expr)
    assert result == op(raw_left, raw_right)



"""File system module."""
# Copyright 2014 Cloudera Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This file may adapt small portions of https://github.com/mtth/hdfs (MIT
# license), see the LICENSES directory.

import posixpath
from functools import wraps as implements

import ibis.common.exceptions as com
from ibis.config import options


class HDFSError(com.IbisError):
    """HDFS Error class."""

    pass


class HDFS:
    """Interface class to HDFS.

    Interface class to HDFS for ibis that abstracts away (and protects
    user/developer against) various 3rd party library API differences.
    """

    def log(self, message: str):
        """Print a log message.

        Parameters
        ----------
        message: string
        """
        print(message)

    def exists(self, path: str) -> bool:
        """Check if the file exists.

        Parameters
        ----------
        path : string

        Returns
        -------
        bool

        Raises
        ------
        NotImplementedError

        """
        raise NotImplementedError

    def status(self, path: str) -> dict:
        """Check if the status of the path.

        Parameters
        ----------
        path : string

        Returns
        -------
        status : dict

        Raises
        ------
        NotImplementedError

        """
        raise NotImplementedError

    def chmod(self, hdfs_path: str, permissions: str):
        """Change permissions of a file of directory.

        Parameters
        ----------
        hdfs_path : string
          Directory or path
        permissions : string
          Octal permissions string

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError

    def chown(self, hdfs_path: str, owner: str = None, group: str = None):
        """Change owner (and/or group) of a file or directory.

        Parameters
        ----------
        hdfs_path : string
          Directory or path
        owner : string, optional
          Name of owner
        group : string, optional
          Name of group

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError

    def head(
        self, hdfs_path: str, nbytes: int = 1024, offset: int = 0
    ) -> bytes:
        """Retrieve the requested number of bytes from a file.

        Parameters
        ----------
        hdfs_path : string
          Absolute HDFS path
        nbytes : int, default 1024 (1K)
          Number of bytes to retrieve
        offset : int, default 0
          Number of bytes at beginning of file to skip before retrieving data

        Returns
        -------
        head_data : bytes

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError

    def get(
        self, hdfs_path: str, local_path: str = '.', overwrite: bool = False
    ) -> str:
        """
        Download remote file or directory to the local filesystem.

        Parameters
        ----------
        hdfs_path : string
        local_path : string, default '.'
        overwrite : bool, default False

        Further keyword arguments passed down to any internal API used.

        Returns
        -------
        written_path : string
          The path to the written file or directory

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError

    def put(
        self,
        hdfs_path: str,
        resource,
        overwrite: bool = False,
        verbose: bool = None,
        **kwargs,
    ) -> str:
        """
        Write file or directory to HDFS.

        Parameters
        ----------
        hdfs_path : string
          Directory or path
        resource : string or buffer-like
          Relative or absolute path to local resource, or a file-like object
        overwrite : boolean, default False
        verbose : boolean, default ibis options.verbose

        Further keyword arguments passed down to any internal API used.

        Returns
        -------
        written_path : string
          The path to the written file or directory

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError

    def put_tarfile(
        self,
        hdfs_path: str,
        local_path: str,
        compression: str = 'gzip',
        verbose: bool = None,
        overwrite: bool = False,
    ):
        """
        Write contents of tar archive to HDFS.

        Write contents of tar archive to HDFS directly without having to
        decompress it locally first.

        Parameters
        ----------
        hdfs_path : string
        local_path : string
        compression : {'gzip', 'bz2', None}
        overwrite : boolean, default False
        verbose : boolean, default None (global default)

        Raises
        ------
        ValueError
            if given compression is none of the following: None, gzip or bz2.
        """
        import tarfile

        modes = {None: 'r', 'gzip': 'r:gz', 'bz2': 'r:bz2'}

        if compression not in modes:
            raise ValueError(
                'Invalid compression type {0}'.format(compression)
            )
        mode = modes[compression]

        tf = tarfile.open(local_path, mode=mode)
        for info in tf:
            if not info.isfile():
                continue

            buf = tf.extractfile(info)
            abspath = posixpath.join(hdfs_path, info.path)
            self.put(abspath, buf, verbose=verbose, overwrite=overwrite)

    def put_zipfile(self, hdfs_path: str, local_path: str):
        """Write contents of zipfile archive to HDFS.

        Parameters
        ----------
        hdfs_path : string
        local_path : string

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError

    def write(
        self,
        hdfs_path: str,
        buf,
        overwrite: bool = False,
        blocksize: int = None,
        replication=None,
        buffersize: int = None,
    ):
        """HDFS Write function.

        Parameters
        ----------
        hdfs_path : string
        buf
        overwrite : bool, defaul False
        blocksize : int
        replication
        buffersize : int

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError

    def mkdir(self, path: str):
        """Create new directory.

        Parameters
        ----------
        path : string
        """
        pass

    def ls(self, hdfs_path: str, status: bool = False) -> list:
        """Return contents of directory.

        Parameters
        ----------
        hdfs_path : string
        status : bool

        Returns
        -------
        list

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError

    def size(self, hdfs_path: str) -> int:
        """Return total size of file or directory.

        Parameters
        ----------
        hdfs_path : basestring

        Returns
        -------
        size : int

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError

    def tail(self, hdfs_path: str, nbytes: int = 1024) -> bytes:
        """Retrieve the requested number of bytes from the end of a file.

        Parameters
        ----------
        hdfs_path : string
        nbytes : int

        Returns
        -------
        data_tail : bytes

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError

    def mv(
        self, hdfs_path_src: str, hdfs_path_dest: str, overwrite: bool = True
    ):
        """Move hdfs_path_src to hdfs_path_dest.

        Parameters
        ----------
        hdfs_path_src: string
        hdfs_path_dest: string
        overwrite : boolean, default True
          Overwrite hdfs_path_dest if it exists.

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError

    def cp(self, hdfs_path_src: str, hdfs_path_dest: str):
        """Copy hdfs_path_src to hdfs_path_dest.

        Parameters
        ----------
        hdfs_path_src : string
        hdfs_path_dest : string

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError

    def rm(self, path: str):
        """Delete a single file.

        Parameters
        ----------
        path : string
        """
        return self.delete(path)

    def rmdir(self, path: str):
        """Delete a directory and all its contents.

        Parameters
        ----------
        path : string
        """
        self.client.delete(path, recursive=True)

    def _find_any_file(self, hdfs_dir):
        contents = self.ls(hdfs_dir, status=True)

        def valid_filename(name):
            head, tail = posixpath.split(name)

            tail = tail.lower()
            return (
                not tail.endswith('.tmp')
                and not tail.endswith('.copying')
                and not tail.startswith('_')
                and not tail.startswith('.')
            )

        for filename, meta in contents:
            if meta['type'].lower() == 'file' and valid_filename(filename):
                return filename
        raise com.IbisError('No files found in the passed directory')


class WebHDFS(HDFS):
    """A WebHDFS-based interface to HDFS using the HDFSCli library."""

    def __init__(self, client):
        self.client = client

    @property
    def protocol(self) -> str:
        """Return the protocol used by WebHDFS.

        Returns
        -------
        protocol : string
        """
        return 'webhdfs'

    def status(self, path: str) -> dict:
        """Retrieve HDFS metadata for path.

        Parameters
        ----------
        path : str

        Returns
        -------
        status : dict
            Client status
        """
        return self.client.status(path)

    @implements(HDFS.chmod)
    def chmod(self, path: str, permissions: str):
        """Change the permissions of a HDFS file.

        Parameters
        ----------
        path : string
        permissions : string
            New octal permissions string of the file.
        """
        self.client.set_permission(path, permissions)

    @implements(HDFS.chown)
    def chown(self, path: str, owner=None, group=None):
        """
        Change the owner of a HDFS file.

        At least one of `owner` and `group` must be specified.

        Parameters
        ----------
        hdfs_path : HDFS path.
        owner : string, optional
        group: string, optional
        """
        self.client.set_owner(path, owner, group)

    @implements(HDFS.exists)
    def exists(self, path: str) -> dict:
        """Check if the HDFS file exists.

        Parameters
        ----------
        path : string

        Returns
        -------
        bool
        """
        return not self.client.status(path, strict=False) is None

    @implements(HDFS.ls)
    def ls(self, hdfs_path: str, status: bool = False) -> list:
        """Return contents of directory.

        Parameters
        ----------
        hdfs_path : string
        status : bool

        Returns
        -------
        list
        """
        return self.client.list(hdfs_path, status=status)

    @implements(HDFS.mkdir)
    def mkdir(self, dir_path: str):
        """Create new directory.

        Parameters
        ----------
        path : string
        """
        self.client.makedirs(dir_path)

    @implements(HDFS.size)
    def size(self, hdfs_path: str) -> int:
        """Return total size of file or directory.

        Parameters
        ----------
        hdfs_path : string

        Returns
        -------
        size : int
        """
        return self.client.content(hdfs_path)['length']

    @implements(HDFS.mv)
    def mv(
        self, hdfs_path_src: str, hdfs_path_dest: str, overwrite: bool = True
    ):
        """Move hdfs_path_src to hdfs_path_dest.

        Parameters
        ----------
        hdfs_path_src: string
        hdfs_path_dest: string
        overwrite : boolean, default True
          Overwrite hdfs_path_dest if it exists.
        """
        if overwrite and self.exists(hdfs_path_dest):
            if self.status(hdfs_path_dest)['type'] == 'FILE':
                self.rm(hdfs_path_dest)
        self.client.rename(hdfs_path_src, hdfs_path_dest)

    def delete(self, hdfs_path: str, recursive: bool = False) -> bool:
        """Delete a file located at `hdfs_path`.

        Parameters
        ----------
        hdfs_path : string
        recursive : bool, default False

        Returns
        -------
        bool
            True if the function was successful.
        """
        return self.client.delete(hdfs_path, recursive=recursive)

    @implements(HDFS.head)
    def head(
        self, hdfs_path: str, nbytes: int = 1024, offset: int = 0
    ) -> bytes:
        """Retrieve the requested number of bytes from a file.

        Parameters
        ----------
        hdfs_path : string
          Absolute HDFS path
        nbytes : int, default 1024 (1K)
          Number of bytes to retrieve
        offset : int, default 0
          Number of bytes at beginning of file to skip before retrieving data

        Returns
        -------
        head_data : bytes
        """
        _reader = self.client.read(hdfs_path, offset=offset, length=nbytes)
        with _reader as reader:
            return reader.read()

    @implements(HDFS.put)
    def put(
        self,
        hdfs_path: str,
        resource,
        overwrite: bool = False,
        verbose: bool = None,
        **kwargs,
    ):
        """
        Write file or directory to HDFS.

        Parameters
        ----------
        hdfs_path : string
          Directory or path
        resource : string or buffer-like
          Relative or absolute path to local resource, or a file-like object
        overwrite : boolean, default False
        verbose : boolean, default ibis options.verbose

        Further keyword arguments passed down to any internal API used.

        Returns
        -------
        written_path : string
          The path to the written file or directory
        """
        verbose = verbose or options.verbose
        if isinstance(resource, str):
            # `resource` is a path.
            return self.client.upload(
                hdfs_path, resource, overwrite=overwrite, **kwargs
            )
        else:
            # `resource` is a file-like object.
            hdfs_path = self.client.resolve(hdfs_path)
            self.client.write(
                hdfs_path, data=resource, overwrite=overwrite, **kwargs
            )
            return hdfs_path

    @implements(HDFS.get)
    def get(
        self,
        hdfs_path: str,
        local_path: str,
        overwrite: bool = False,
        verbose: bool = None,
        **kwargs,
    ) -> str:
        """
        Download remote file or directory to the local filesystem.

        Parameters
        ----------
        hdfs_path : string
        local_path : string, default '.'
        overwrite : bool, default False

        Further keyword arguments passed down to any internal API used.

        Returns
        -------
        written_path : string
          The path to the written file or directory
        """
        verbose = verbose or options.verbose
        return self.client.download(
            hdfs_path, local_path, overwrite=overwrite, **kwargs
        )



import collections

from multipledispatch import Dispatcher

import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.util as util


class Schema:

    """An object for holding table schema information, i.e., column names and
    types.

    Parameters
    ----------
    names : Sequence[str]
        A sequence of ``str`` indicating the name of each column.
    types : Sequence[DataType]
        A sequence of :class:`ibis.expr.datatypes.DataType` objects
        representing type of each column.
    """

    __slots__ = 'names', 'types', '_name_locs'

    def __init__(self, names, types):
        if not isinstance(names, list):
            names = list(names)

        self.names = names
        self.types = list(map(dt.dtype, types))

        self._name_locs = dict((v, i) for i, v in enumerate(self.names))

        if len(self._name_locs) < len(self.names):
            duplicate_names = list(self.names)
            for v in self._name_locs.keys():
                duplicate_names.remove(v)
            raise com.IntegrityError(
                'Duplicate column name(s): {}'.format(duplicate_names)
            )

    def __repr__(self):
        space = 2 + max(map(len, self.names), default=0)
        return "ibis.Schema {{{}\n}}".format(
            util.indent(
                ''.join(
                    '\n{}{}'.format(name.ljust(space), str(type))
                    for name, type in zip(self.names, self.types)
                ),
                2,
            )
        )

    def __hash__(self):
        return hash((type(self), tuple(self.names), tuple(self.types)))

    def __len__(self):
        return len(self.names)

    def __iter__(self):
        return iter(self.names)

    def __contains__(self, name):
        return name in self._name_locs

    def __getitem__(self, name):
        return self.types[self._name_locs[name]]

    def __getstate__(self):
        return {slot: getattr(self, slot) for slot in self.__class__.__slots__}

    def __setstate__(self, instance_dict):
        for key, value in instance_dict.items():
            setattr(self, key, value)

    def delete(self, names_to_delete):
        for name in names_to_delete:
            if name not in self:
                raise KeyError(name)

        new_names, new_types = [], []
        for name, type_ in zip(self.names, self.types):
            if name in names_to_delete:
                continue
            new_names.append(name)
            new_types.append(type_)

        return Schema(new_names, new_types)

    @classmethod
    def from_tuples(cls, values):
        if not isinstance(values, (list, tuple)):
            values = list(values)

        names, types = zip(*values) if values else ([], [])
        return Schema(names, types)

    @classmethod
    def from_dict(cls, dictionary):
        return Schema(*zip(*dictionary.items()))

    def equals(self, other, cache=None):
        return self.names == other.names and self.types == other.types

    def __eq__(self, other):
        return self.equals(other)

    def __gt__(self, other):
        return set(self.items()) > set(other.items())

    def __ge__(self, other):
        return set(self.items()) >= set(other.items())

    def append(self, schema):
        return Schema(self.names + schema.names, self.types + schema.types)

    def items(self):
        return zip(self.names, self.types)

    def name_at_position(self, i):
        """
        """
        upper = len(self.names) - 1
        if not 0 <= i <= upper:
            raise ValueError(
                'Column index must be between 0 and {:d}, inclusive'.format(
                    upper
                )
            )
        return self.names[i]


class HasSchema:

    """
    Base class representing a structured dataset with a well-defined
    schema.

    Base implementation is for tables that do not reference a particular
    concrete dataset or database table.
    """

    def __repr__(self):
        return '{}({})'.format(type(self).__name__, repr(self.schema))

    def has_schema(self):
        return True

    def equals(self, other, cache=None):
        return type(self) == type(other) and self.schema.equals(
            other.schema, cache=cache
        )

    def root_tables(self):
        return [self]

    @property
    def schema(self):
        raise NotImplementedError


schema = Dispatcher('schema')
infer = Dispatcher('infer')


@schema.register(Schema)
def identity(s):
    return s


@schema.register(collections.abc.Mapping)
def schema_from_mapping(d):
    return Schema.from_dict(d)


@schema.register(collections.abc.Iterable)
def schema_from_pairs(lst):
    return Schema.from_tuples(lst)


@schema.register(collections.abc.Iterable, collections.abc.Iterable)
def schema_from_names_types(names, types):
    return Schema(names, types)



import pytest

import ibis.expr.datatypes as dt
from ibis.tests.backends import Pandas, PySpark
from ibis.udf.vectorized import elementwise


@elementwise(input_type=[dt.double], output_type=dt.double)
def add_one(s):
    return s + 1


@pytest.mark.only_on_backends([Pandas, PySpark])
@pytest.mark.xfail_unsupported
def test_elementwise_udf(backend, alltypes, df):
    result = add_one(alltypes['double_col']).execute()
    expected = add_one.func(df['double_col'])
    backend.assert_series_equal(result, expected, check_names=False)



"""The pandas backend is a departure from the typical ibis backend in that it
doesn't compile to anything, and the execution of the ibis expression
is under the purview of ibis itself rather than executing SQL on a server.

Design
------
The pandas backend uses a technique called `multiple dispatch
<https://en.wikipedia.org/wiki/Multiple_dispatch>`_, implemented in a
third-party open source library called `multipledispatch
<https://github.com/mrocklin/multipledispatch>`_.

Multiple dispatch is a generalization of standard single-dispatch runtime
polymorphism to multiple arguments.

Compilation
-----------
This is a no-op because we execute ibis expressions directly.

Execution
---------
Execution is divided into different dispatched functions, each arising from
a different use case.

A top level function `execute` exists to provide the API for executing an ibis
expression against in-memory data.

The general flow of execution is:

::
       If the current operation is in scope:
           return it
       Else:
           execute the arguments of the current node

       execute the current node with its executed arguments

Specifically, execute is comprised of a series of steps that happen at
different times during the loop.

1. ``pre_execute``
------------------
First, at the beginning of the main execution loop, ``pre_execute`` is called.
This function serves a similar purpose to ``data_preload``, the key difference
being that ``pre_execute`` is called *every time* there's a call to execute.

By default this function does nothing.

2. ``execute_node``
-------------------

Second, when an expression is ready to be evaluated we call
:func:`~ibis.pandas.core.execute` on the expressions arguments and then
:func:`~ibis.pandas.dispatch.execute_node` on the expression with its
now-materialized arguments.

3. ``post_execute``
-------------------
The final step--``post_execute``--is called immediately after the previous call
to ``execute_node`` and takes the instance of the
:class:`~ibis.expr.operations.Node` just computed and the result of the
computation.

The purpose of this function is to allow additional computation to happen in
the context of the current level of the execution loop. You might be wondering
That may sound vague, so let's look at an example.

Let's say you want to take a three day rolling average, and you want to include
3 days of data prior to the first date of the input. You don't want to see that
data in the result for a few reasons, one of which is that it would break the
contract of window functions: given N rows of input there are N rows of output.

Defining a ``post_execute`` rule for :class:`~ibis.expr.operations.WindowOp`
allows you to encode such logic. One might want to implement this using
:class:`~ibis.expr.operations.ScalarParameter`, in which case the ``scope``
passed to ``post_execute`` would be the bound values passed in at the time the
``execute`` method was called.
"""

from __future__ import absolute_import

import datetime
import functools
import numbers

import numpy as np
import pandas as pd
import toolz
from multipledispatch import Dispatcher

import ibis
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.types as ir
import ibis.expr.window as win
import ibis.pandas.aggcontext as agg_ctx
from ibis.client import find_backends
from ibis.pandas.dispatch import (
    execute_literal,
    execute_node,
    post_execute,
    pre_execute,
)

integer_types = np.integer, int
floating_types = (numbers.Real,)
numeric_types = integer_types + floating_types
boolean_types = bool, np.bool_
fixed_width_types = numeric_types + boolean_types
date_types = (datetime.date,)
time_types = (datetime.time,)
timestamp_types = pd.Timestamp, datetime.datetime, np.datetime64
timedelta_types = pd.Timedelta, datetime.timedelta, np.timedelta64
temporal_types = date_types + time_types + timestamp_types + timedelta_types
scalar_types = fixed_width_types + temporal_types
simple_types = scalar_types + (str, type(None))


@functools.singledispatch
def is_computable_input(arg):
    """All inputs are not computable without a specific override."""
    return False


@is_computable_input.register(ibis.client.Client)
@is_computable_input.register(ir.Expr)
@is_computable_input.register(dt.DataType)
@is_computable_input.register(type(None))
@is_computable_input.register(win.Window)
@is_computable_input.register(tuple)
def is_computable_input_arg(arg):
    """Return whether `arg` is a valid computable argument."""
    return True


# Register is_computable_input for each scalar type (int, float, date, etc).
# We use consume here to avoid leaking the iteration variable into the module.
ibis.util.consume(
    is_computable_input.register(t)(is_computable_input_arg)
    for t in scalar_types
)


def execute_with_scope(expr, scope, aggcontext=None, clients=None, **kwargs):
    """Execute an expression `expr`, with data provided in `scope`.

    Parameters
    ----------
    expr : ibis.expr.types.Expr
        The expression to execute.
    scope : collections.Mapping
        A dictionary mapping :class:`~ibis.expr.operations.Node` subclass
        instances to concrete data such as a pandas DataFrame.
    aggcontext : Optional[ibis.pandas.aggcontext.AggregationContext]

    Returns
    -------
    result : scalar, pd.Series, pd.DataFrame
    """
    op = expr.op()

    # Call pre_execute, to allow clients to intercept the expression before
    # computing anything *and* before associating leaf nodes with data. This
    # allows clients to provide their own data for each leaf.
    if clients is None:
        clients = list(find_backends(expr))

    if aggcontext is None:
        aggcontext = agg_ctx.Summarize()

    pre_executed_scope = pre_execute(
        op, *clients, scope=scope, aggcontext=aggcontext, **kwargs
    )
    new_scope = toolz.merge(scope, pre_executed_scope)
    result = execute_until_in_scope(
        expr,
        new_scope,
        aggcontext=aggcontext,
        clients=clients,
        # XXX: we *explicitly* pass in scope and not new_scope here so that
        # post_execute sees the scope of execute_with_scope, not the scope of
        # execute_until_in_scope
        post_execute_=functools.partial(
            post_execute,
            scope=scope,
            aggcontext=aggcontext,
            clients=clients,
            **kwargs,
        ),
        **kwargs,
    )[op]

    return result


def execute_until_in_scope(
    expr, scope, aggcontext=None, clients=None, post_execute_=None, **kwargs
):
    """Execute until our op is in `scope`.

    Parameters
    ----------
    expr : ibis.expr.types.Expr
    scope : Mapping
    aggcontext : Optional[AggregationContext]
    clients : List[ibis.client.Client]
    kwargs : Mapping
    """
    # these should never be None
    assert aggcontext is not None, 'aggcontext is None'
    assert clients is not None, 'clients is None'
    assert post_execute_ is not None, 'post_execute_ is None'

    # base case: our op has been computed (or is a leaf data node), so
    # return the corresponding value
    op = expr.op()
    if op in scope:
        return scope
    elif isinstance(op, ops.Literal):
        # special case literals to avoid the overhead of dispatching
        # execute_node
        return {
            op: execute_literal(
                op, op.value, expr.type(), aggcontext=aggcontext, **kwargs
            )
        }

    pre_executed_scope = pre_execute(
        op, *clients, scope=scope, aggcontext=aggcontext, **kwargs
    )
    new_scope = toolz.merge(scope, pre_executed_scope)

    # Short circuit: if pre_execute puts op in scope, then we don't need to
    # execute its computable_args
    if op in new_scope:
        return new_scope

    # figure out what arguments we're able to compute on based on the
    # expressions inputs. things like expressions, None, and scalar types are
    # computable whereas ``list``s are not
    computable_args = [arg for arg in op.inputs if is_computable_input(arg)]

    # recursively compute each node's arguments until we've changed type
    scopes = [
        execute_until_in_scope(
            arg,
            new_scope,
            aggcontext=aggcontext,
            post_execute_=post_execute_,
            clients=clients,
            **kwargs,
        )
        if hasattr(arg, 'op')
        else {arg: arg}
        for arg in computable_args
    ]

    # if we're unable to find data then raise an exception
    if not scopes and computable_args:
        raise com.UnboundExpressionError(
            'Unable to find data for expression:\n{}'.format(repr(expr))
        )

    # there should be exactly one dictionary per computable argument
    assert len(computable_args) == len(scopes)

    new_scope = toolz.merge(new_scope, *scopes)

    # pass our computed arguments to this node's execute_node implementation
    data = [
        new_scope[arg.op()] if hasattr(arg, 'op') else arg
        for arg in computable_args
    ]
    result = execute_node(
        op,
        *data,
        scope=scope,
        aggcontext=aggcontext,
        clients=clients,
        **kwargs,
    )
    computed = post_execute_(op, result)
    return {op: computed}


execute = Dispatcher('execute')


@execute.register(ir.Expr)
def main_execute(expr, params=None, scope=None, aggcontext=None, **kwargs):
    """Execute an expression against data that are bound to it. If no data
    are bound, raise an Exception.

    Parameters
    ----------
    expr : ibis.expr.types.Expr
        The expression to execute
    params : Mapping[ibis.expr.types.Expr, object]
        The data that an unbound parameter in `expr` maps to
    scope : Mapping[ibis.expr.operations.Node, object]
        Additional scope, mapping ibis operations to data
    aggcontext : Optional[ibis.pandas.aggcontext.AggregationContext]
        An object indicating how to compute aggregations. For example,
        a rolling mean needs to be computed differently than the mean of a
        column.
    kwargs : Dict[str, object]
        Additional arguments that can potentially be used by individual node
        execution

    Returns
    -------
    result : Union[
        pandas.Series, pandas.DataFrame, ibis.pandas.core.simple_types
    ]

    Raises
    ------
    ValueError
        * If no data are bound to the input expression
    """
    if scope is None:
        scope = {}

    if params is None:
        params = {}

    # TODO: make expresions hashable so that we can get rid of these .op()
    # calls everywhere
    params = {k.op() if hasattr(k, 'op') else k: v for k, v in params.items()}

    new_scope = toolz.merge(scope, params)
    return execute_with_scope(expr, new_scope, aggcontext=aggcontext, **kwargs)


def execute_and_reset(
    expr, params=None, scope=None, aggcontext=None, **kwargs
):
    """Execute an expression against data that are bound to it. If no data
    are bound, raise an Exception.

    Notes
    -----
    The difference between this function and :func:`~ibis.pandas.core.execute`
    is that this function resets the index of the result, if the result has
    an index.

    Parameters
    ----------
    expr : ibis.expr.types.Expr
        The expression to execute
    params : Mapping[ibis.expr.types.Expr, object]
        The data that an unbound parameter in `expr` maps to
    scope : Mapping[ibis.expr.operations.Node, object]
        Additional scope, mapping ibis operations to data
    aggcontext : Optional[ibis.pandas.aggcontext.AggregationContext]
        An object indicating how to compute aggregations. For example,
        a rolling mean needs to be computed differently than the mean of a
        column.
    kwargs : Dict[str, object]
        Additional arguments that can potentially be used by individual node
        execution

    Returns
    -------
    result : Union[
        pandas.Series, pandas.DataFrame, ibis.pandas.core.simple_types
    ]

    Raises
    ------
    ValueError
        * If no data are bound to the input expression
    """
    result = execute(
        expr, params=params, scope=scope, aggcontext=aggcontext, **kwargs
    )
    if isinstance(result, pd.DataFrame):
        schema = expr.schema()
        df = result.reset_index()
        return df.loc[:, schema.names]
    elif isinstance(result, pd.Series):
        return result.reset_index(drop=True)
    return result



print("import: 'ibis'")
import ibis

print("import: 'ibis.expr'")
import ibis.expr

print("import: 'ibis.expr.tests'")
import ibis.expr.tests

print("import: 'ibis.hive'")
import ibis.hive

print("import: 'ibis.hive.tests'")
import ibis.hive.tests

print("import: 'ibis.impala'")
import ibis.impala

print("import: 'ibis.impala.tests'")
import ibis.impala.tests

print("import: 'ibis.pandas'")
import ibis.pandas

print("import: 'ibis.pandas.tests'")
import ibis.pandas.tests

print("import: 'ibis.spark'")
import ibis.spark

print("import: 'ibis.spark.tests'")
import ibis.spark.tests

print("import: 'ibis.sql'")
import ibis.sql

print("import: 'ibis.sql.postgres'")
import ibis.sql.postgres

print("import: 'ibis.sql.postgres.tests'")
import ibis.sql.postgres.tests

print("import: 'ibis.sql.presto'")
import ibis.sql.presto

print("import: 'ibis.sql.presto.tests'")
import ibis.sql.presto.tests

print("import: 'ibis.sql.redshift'")
import ibis.sql.redshift

print("import: 'ibis.sql.redshift.tests'")
import ibis.sql.redshift.tests

print("import: 'ibis.sql.sqlite'")
import ibis.sql.sqlite

print("import: 'ibis.sql.sqlite.tests'")
import ibis.sql.sqlite.tests

print("import: 'ibis.sql.tests'")
import ibis.sql.tests

print("import: 'ibis.sql.vertica'")
import ibis.sql.vertica

print("import: 'ibis.sql.vertica.tests'")
import ibis.sql.vertica.tests

print("import: 'ibis.tests'")
import ibis.tests




import pytest
from pandas import Timestamp

import ibis
from ibis import literal as L

pytest.importorskip('clickhouse_driver')
pytestmark = pytest.mark.clickhouse


@pytest.mark.parametrize(
    'expr',
    [
        L(Timestamp('2015-01-01 12:34:56')),
        L(Timestamp('2015-01-01 12:34:56').to_pydatetime()),
        ibis.timestamp('2015-01-01 12:34:56'),
    ],
)
def test_timestamp_literals(con, translate, expr):
    expected = "toDateTime('2015-01-01 12:34:56')"

    assert translate(expr) == expected
    assert con.execute(expr) == Timestamp('2015-01-01 12:34:56')


@pytest.mark.parametrize(
    ('value', 'expected'),
    [
        ('simple', "'simple'"),
        ('I can\'t', "'I can\\'t'"),
        ('An "escape"', "'An \"escape\"'"),
    ],
)
def test_string_literals(con, translate, value, expected):
    expr = ibis.literal(value)
    assert translate(expr) == expected
    # TODO clickhouse-driver escaping problem
    # assert con.execute(expr) == expected


@pytest.mark.parametrize(('value', 'expected'), [(5, '5'), (1.5, '1.5')])
def test_number_literals(con, translate, value, expected):
    expr = ibis.literal(value)
    assert translate(expr) == expected
    assert con.execute(expr) == value


@pytest.mark.parametrize(('value', 'expected'), [(True, '1'), (False, '0')])
def test_boolean_literals(con, translate, value, expected):
    expr = ibis.literal(value)
    assert translate(expr) == expected
    assert con.execute(expr) == value



import datetime
from operator import methodcaller

import numpy as np
import pandas as pd
import pandas.util.testing as tm  # noqa: E402
import pytest
from pkg_resources import parse_version
from pytest import param

import ibis
from ibis import literal as L  # noqa: E402
from ibis.expr import datatypes as dt

pytestmark = pytest.mark.pandas


@pytest.mark.parametrize(
    ('case_func', 'expected_func'),
    [
        (lambda v: v.strftime('%Y%m%d'), lambda vt: vt.strftime('%Y%m%d')),
        (lambda v: v.year(), lambda vt: vt.year),
        (lambda v: v.month(), lambda vt: vt.month),
        (lambda v: v.day(), lambda vt: vt.day),
        (lambda v: v.hour(), lambda vt: vt.hour),
        (lambda v: v.minute(), lambda vt: vt.minute),
        (lambda v: v.second(), lambda vt: vt.second),
        (lambda v: v.millisecond(), lambda vt: int(vt.microsecond / 1e3)),
    ]
    + [
        (methodcaller('strftime', pattern), methodcaller('strftime', pattern))
        for pattern in [
            '%Y%m%d %H',
            'DD BAR %w FOO "DD"',
            'DD BAR %w FOO "D',
            'DD BAR "%w" FOO "D',
            'DD BAR "%d" FOO "D',
            'DD BAR "%c" FOO "D',
            'DD BAR "%x" FOO "D',
            'DD BAR "%X" FOO "D',
        ]
    ],
)
def test_timestamp_functions(case_func, expected_func):
    v = L('2015-09-01 14:48:05.359').cast('timestamp')
    vt = datetime.datetime(
        year=2015,
        month=9,
        day=1,
        hour=14,
        minute=48,
        second=5,
        microsecond=359000,
    )
    result = case_func(v)
    expected = expected_func(vt)
    assert ibis.pandas.execute(result) == expected


@pytest.mark.parametrize(
    'column',
    ['datetime_strings_naive', 'datetime_strings_ny', 'datetime_strings_utc'],
)
def test_cast_datetime_strings_to_date(t, df, column):
    expr = t[column].cast('date')
    result = expr.execute()
    expected = (
        pd.to_datetime(df[column], infer_datetime_format=True)
        .dt.normalize()
        .dt.tz_localize(None)
    )
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize(
    'column',
    ['datetime_strings_naive', 'datetime_strings_ny', 'datetime_strings_utc'],
)
def test_cast_datetime_strings_to_timestamp(t, df, column):
    expr = t[column].cast('timestamp')
    result = expr.execute()
    expected = pd.to_datetime(df[column], infer_datetime_format=True)
    if getattr(expected.dtype, 'tz', None) is not None:
        expected = expected.dt.tz_convert(None)
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize(
    'column',
    ['plain_datetimes_naive', 'plain_datetimes_ny', 'plain_datetimes_utc'],
)
def test_cast_integer_to_temporal_type(t, df, column):
    column_type = t[column].type()
    expr = t.plain_int64.cast(column_type)
    result = expr.execute()
    expected = pd.Series(
        pd.to_datetime(df.plain_int64.values, unit='ns').values,
        index=df.index,
        name='plain_int64',
    ).dt.tz_localize(column_type.timezone)
    tm.assert_series_equal(result, expected)


def test_cast_integer_to_date(t, df):
    expr = t.plain_int64.cast('date')
    result = expr.execute()
    expected = pd.Series(
        pd.to_datetime(df.plain_int64.values, unit='D').values,
        index=df.index,
        name='plain_int64',
    )
    tm.assert_series_equal(result, expected)


def test_times_ops(t, df):
    result = t.plain_datetimes_naive.time().between('10:00', '10:00').execute()
    expected = pd.Series(np.zeros(len(df), dtype=bool))
    tm.assert_series_equal(result, expected)

    result = t.plain_datetimes_naive.time().between('01:00', '02:00').execute()
    expected = pd.Series(np.ones(len(df), dtype=bool))
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ('tz', 'rconstruct'),
    [('US/Eastern', np.zeros), ('UTC', np.ones), (None, np.ones)],
)
@pytest.mark.parametrize(
    'column', ['plain_datetimes_utc', 'plain_datetimes_naive']
)
def test_times_ops_with_tz(t, df, tz, rconstruct, column):
    expected = pd.Series(rconstruct(len(df), dtype=bool))
    time = t[column].time()
    expr = time.between('01:00', '02:00', timezone=tz)
    result = expr.execute()
    tm.assert_series_equal(result, expected)

    # Test that casting behavior is the same as using the timezone kwarg
    ts = t[column].cast(dt.Timestamp(timezone=tz))
    expr = ts.time().between('01:00', '02:00')
    result = expr.execute()
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ('op', 'expected'),
    [
        param(lambda x, y: x + y, lambda x, y: x.values * 2, id='add'),
        param(lambda x, y: x - y, lambda x, y: x.values - y.values, id='sub'),
        param(lambda x, y: x * 2, lambda x, y: x.values * 2, id='mul'),
        param(
            lambda x, y: x // 2,
            lambda x, y: x.values // 2,
            id='floordiv',
            marks=pytest.mark.xfail(
                parse_version(pd.__version__) < parse_version('0.23.0'),
                raises=TypeError,
                reason=(
                    'pandas versions less than 0.23.0 do not support floor '
                    'division involving timedelta columns'
                ),
            ),
        ),
    ],
)
def test_interval_arithmetic(op, expected):
    data = pd.timedelta_range('0 days', '10 days', freq='D')
    con = ibis.pandas.connect(
        {'df1': pd.DataFrame({'td': data}), 'df2': pd.DataFrame({'td': data})}
    )
    t1 = con.table('df1')
    expr = op(t1.td, t1.td)
    result = expr.execute()
    expected = pd.Series(expected(data, data), name='td')
    tm.assert_series_equal(result, expected)



from __future__ import absolute_import

from functools import partial

import toolz
from multipledispatch import Dispatcher

import ibis
import ibis.common.exceptions as com
import ibis.expr.operations as ops

# Individual operation execution
execute_node = Dispatcher(
    'execute_node',
    doc=(
        'Execute an individual operation given the operation and its computed '
        'arguments'
    ),
)


@execute_node.register(ops.Node)
def execute_node_without_scope(node, **kwargs):
    raise com.UnboundExpressionError(
        (
            'Node of type {!r} has no data bound to it. '
            'You probably tried to execute an expression without a data '
            'source.'
        ).format(type(node).__name__)
    )


pre_execute = Dispatcher(
    'pre_execute',
    doc="""\
Given a node, compute a (possibly partial) scope prior to standard execution.

Notes
-----
This function is useful if parts of the tree structure need to be executed at
the same time or if there are other reasons to need to interrupt the regular
depth-first traversal of the tree.
""",
)


# Default returns an empty scope
@pre_execute.register(ops.Node)
@pre_execute.register(ops.Node, ibis.client.Client)
def pre_execute_default(node, *clients, **kwargs):
    return {}


# Merge the results of all client pre-execution with scope
@pre_execute.register(ops.Node, [ibis.client.Client])
def pre_execute_multiple_clients(node, *clients, scope=None, **kwargs):
    return toolz.merge(
        scope, *map(partial(pre_execute, node, scope=scope, **kwargs), clients)
    )


execute_literal = Dispatcher(
    'execute_literal',
    doc="""\
Special case literal execution to avoid the dispatching overhead of
``execute_node``.

Parameters
----------
op : ibis.expr.operations.Node
value : object
    The literal value of the object, e.g., int, float.
datatype : ibis.expr.datatypes.DataType
    Used to specialize on expressions whose underlying value is of a different
    type than its would-be type. For example, interval values are represented
    by an integer.
""",
)


post_execute = Dispatcher(
    'post_execute',
    doc="""\
Execute code on the result of a computation.

Parameters
----------
op : ibis.expr.operations.Node
    The operation that was just executed
data : object
    The result of the computation
""",
)


@post_execute.register(ops.Node, object)
def post_execute_default(op, data, **kwargs):
    return data


execute = Dispatcher("execute")



import inspect
from collections import OrderedDict

import ibis.expr.rules as rlz
import ibis.util as util

try:
    from cytoolz import unique
except ImportError:
    from toolz import unique


_undefined = object()  # marker for missing argument


class Argument:
    """Argument definition."""

    __slots__ = 'validator', 'default', 'show'

    def __init__(self, validator, default=_undefined, show=True):
        """Argument constructor

        Parameters
        ----------
        validator : Union[Callable[[arg], coerced], Type, Tuple[Type]]
            Function which handles validation and/or coercion of the given
            argument.
        default : Union[Any, Callable[[], str]]
            In case of missing (None) value for validation this will be used.
            Note, that default value (except for None) must also pass the inner
            validator.
            If callable is passed, it will be executed just before the inner,
            and itsreturn value will be treaded as default.
        show : bool
            Whether to show this argument in an :class:`~ibis.expr.types.Expr`
            that contains it.
        """
        self.default = default
        self.show = show
        if isinstance(validator, type):
            self.validator = rlz.instance_of(validator)
        elif isinstance(validator, tuple):
            assert util.all_of(validator, type)
            self.validator = rlz.instance_of(validator)
        elif callable(validator):
            self.validator = validator
        else:
            raise TypeError(
                'Argument validator must be a callable, type or '
                'tuple of types, given: {}'.format(validator)
            )

    def __eq__(self, other):
        return (
            self.validator == other.validator and self.default == other.default
        )

    @property
    def optional(self):
        return self.default is not _undefined

    def validate(self, value=_undefined, name=None):
        """
        Parameters
        ----------
        value : Any, default undefined
          Raises TypeError if argument is mandatory but not value has been
          given.
        name : Optional[str]
          Argument name for error message
        """
        if self.optional:
            if value is _undefined or value is None:
                if self.default is None:
                    return None
                elif util.is_function(self.default):
                    value = self.default()
                else:
                    value = self.default
        elif value is _undefined:
            if name is not None:
                name_msg = "argument `{}`".format(name)
            else:
                name_msg = "unnamed argument"
            raise TypeError("Missing required value for {}".format(name_msg))

        return self.validator(value)

    __call__ = validate  # syntactic sugar


class TypeSignature(OrderedDict):

    __slots__ = ()

    @classmethod
    def from_dtypes(cls, dtypes):
        return cls(
            ('_{}'.format(i), Argument(rlz.value(dtype)))
            for i, dtype in enumerate(dtypes)
        )

    def validate(self, *args, **kwargs):
        parameters = [
            inspect.Parameter(
                name,
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                default=_undefined,
            )
            for (name, argument) in self.items()
        ]
        sig = inspect.Signature(parameters)
        bindings = sig.bind(*args, **kwargs)

        # The inspect.Parameter objects in parameters all have default
        # value _undefined, which will be bound to all arguments that weren't
        # passed in.
        bindings.apply_defaults()

        result = []
        for (name, arg_value) in bindings.arguments.items():
            argument = self[name]
            # If this arg wasn't passed in: since argument.default has the
            # correct value and _undefined was given as the default for the
            # Parameter object corresponding to this argument, arg_value got
            # the value _undefined when bindings.apply_defaults() was called,
            # so the behavior of argument.validate here is correct.
            value = argument.validate(arg_value, name=name)
            result.append((name, value))

        return result

    __call__ = validate  # syntactic sugar

    def names(self):
        return tuple(self.keys())


class AnnotableMeta(type):
    @classmethod
    def __prepare__(metacls, name, bases, **kwds):
        return OrderedDict()

    def __new__(meta, name, bases, dct):
        slots, signature = [], TypeSignature()

        for parent in bases:
            # inherit parent slots
            if hasattr(parent, '__slots__'):
                slots += parent.__slots__
            # inherit from parent signatures
            if hasattr(parent, 'signature'):
                signature.update(parent.signature)

        # finally apply definitions from the currently created class
        # thanks to __prepare__ attrs are already ordered
        attribs = {}
        for k, v in dct.items():
            if isinstance(v, Argument):
                # so we can set directly
                signature[k] = v
            else:
                attribs[k] = v

        # if slots or signature are defined no inheritance happens
        signature = attribs.get('signature', signature)
        slots = attribs.get('__slots__', tuple(slots)) + signature.names()

        attribs['signature'] = signature
        attribs['__slots__'] = tuple(unique(slots))

        return super().__new__(meta, name, bases, attribs)


class Annotable(metaclass=AnnotableMeta):

    __slots__ = ()

    def __init__(self, *args, **kwargs):
        for name, value in self.signature.validate(*args, **kwargs):
            setattr(self, name, value)
        self._validate()

    def _validate(self):
        pass

    @property
    def args(self):
        return tuple(getattr(self, name) for name in self.signature.names())

    @property
    def argnames(self):
        return self.signature.names()



import pytest
from pytest import param

import ibis
import ibis.expr.datatypes as dt
from ibis.tests.backends import Clickhouse, Impala, PostgreSQL, PySpark, Spark


def test_string_col_is_unicode(backend, alltypes, df):
    dtype = alltypes.string_col.type()
    assert dtype == dt.String(nullable=dtype.nullable)
    is_text_type = lambda x: isinstance(x, str)  # noqa: E731
    assert df.string_col.map(is_text_type).all()
    result = alltypes.string_col.execute()
    assert result.map(is_text_type).all()


@pytest.mark.parametrize(
    ('result_func', 'expected_func'),
    [
        param(
            lambda t: t.string_col.contains('6'),
            lambda t: t.string_col.str.contains('6'),
            id='contains',
        ),
        param(
            lambda t: t.string_col.like('6%'),
            lambda t: t.string_col.str.contains('6.*'),
            id='like',
        ),
        param(
            lambda t: t.string_col.like('6^%'),
            lambda t: t.string_col.str.contains('6%'),
            id='complex_like_escape',
        ),
        param(
            lambda t: t.string_col.like('6^%%'),
            lambda t: t.string_col.str.contains('6%.*'),
            id='complex_like_escape_match',
        ),
        param(
            lambda t: t.string_col.re_search(r'[[:digit:]]+'),
            lambda t: t.string_col.str.contains(r'\d+'),
            id='re_search',
            marks=pytest.mark.xfail_backends((Spark, PySpark)),
        ),
        param(
            lambda t: t.string_col.re_extract(r'([[:digit:]]+)', 0),
            lambda t: t.string_col.str.extract(r'(\d+)', expand=False),
            id='re_extract',
            marks=pytest.mark.xfail_backends((Spark, PySpark)),
        ),
        param(
            lambda t: t.string_col.re_replace(r'[[:digit:]]+', 'a'),
            lambda t: t.string_col.str.replace(r'\d+', 'a'),
            id='re_replace',
            marks=pytest.mark.xfail_backends((Spark, PySpark)),
        ),
        param(
            lambda t: t.string_col.re_search(r'\\d+'),
            lambda t: t.string_col.str.contains(r'\d+'),
            id='re_search_spark',
            marks=pytest.mark.xpass_backends((Clickhouse, Impala, Spark)),
        ),
        param(
            lambda t: t.string_col.re_extract(r'(\\d+)', 0),
            lambda t: t.string_col.str.extract(r'(\d+)', expand=False),
            id='re_extract_spark',
            marks=pytest.mark.xpass_backends((Clickhouse, Impala, Spark)),
        ),
        param(
            lambda t: t.string_col.re_replace(r'\\d+', 'a'),
            lambda t: t.string_col.str.replace(r'\d+', 'a'),
            id='re_replace_spark',
            marks=pytest.mark.xpass_backends((Clickhouse, Impala, Spark)),
        ),
        param(
            lambda t: t.string_col.re_search(r'\d+'),
            lambda t: t.string_col.str.contains(r'\d+'),
            id='re_search_spark',
            marks=pytest.mark.xfail_backends((Clickhouse, Impala, Spark)),
        ),
        param(
            lambda t: t.string_col.re_extract(r'(\d+)', 0),
            lambda t: t.string_col.str.extract(r'(\d+)', expand=False),
            id='re_extract_spark',
            marks=pytest.mark.xfail_backends((Clickhouse, Impala, Spark)),
        ),
        param(
            lambda t: t.string_col.re_replace(r'\d+', 'a'),
            lambda t: t.string_col.str.replace(r'\d+', 'a'),
            id='re_replace_spark',
            marks=pytest.mark.xfail_backends((Clickhouse, Impala, Spark)),
        ),
        param(
            lambda t: t.string_col.repeat(2),
            lambda t: t.string_col * 2,
            id='repeat',
        ),
        param(
            lambda t: t.string_col.translate('0', 'a'),
            lambda t: t.string_col.str.translate(str.maketrans('0', 'a')),
            id='translate',
        ),
        param(
            lambda t: t.string_col.find('a'),
            lambda t: t.string_col.str.find('a'),
            id='find',
        ),
        param(
            lambda t: t.string_col.lpad(10, 'a'),
            lambda t: t.string_col.str.pad(10, fillchar='a', side='left'),
            id='lpad',
        ),
        param(
            lambda t: t.string_col.rpad(10, 'a'),
            lambda t: t.string_col.str.pad(10, fillchar='a', side='right'),
            id='rpad',
        ),
        param(
            lambda t: t.string_col.find_in_set(['1']),
            lambda t: t.string_col.str.find('1'),
            id='find_in_set',
        ),
        param(
            lambda t: t.string_col.find_in_set(['a']),
            lambda t: t.string_col.str.find('a'),
            id='find_in_set_all_missing',
        ),
        param(
            lambda t: t.string_col.lower(),
            lambda t: t.string_col.str.lower(),
            id='lower',
        ),
        param(
            lambda t: t.string_col.upper(),
            lambda t: t.string_col.str.upper(),
            id='upper',
        ),
        param(
            lambda t: t.string_col.reverse(),
            lambda t: t.string_col.str[::-1],
            id='reverse',
        ),
        param(
            lambda t: t.string_col.ascii_str(),
            lambda t: t.string_col.map(ord).astype('int32'),
            id='ascii_str',
        ),
        param(
            lambda t: t.string_col.length(),
            lambda t: t.string_col.str.len().astype('int32'),
            id='length',
        ),
        param(
            lambda t: t.string_col.strip(),
            lambda t: t.string_col.str.strip(),
            id='strip',
        ),
        param(
            lambda t: t.string_col.lstrip(),
            lambda t: t.string_col.str.lstrip(),
            id='lstrip',
        ),
        param(
            lambda t: t.string_col.rstrip(),
            lambda t: t.string_col.str.rstrip(),
            id='rstrip',
        ),
        param(
            lambda t: t.string_col.capitalize(),
            lambda t: t.string_col.str.capitalize(),
            id='capitalize',
        ),
        param(
            lambda t: t.date_string_col.substr(2, 3),
            lambda t: t.date_string_col.str[2:5],
            id='substr',
        ),
        param(
            lambda t: t.date_string_col.left(2),
            lambda t: t.date_string_col.str[:2],
            id='left',
        ),
        param(
            lambda t: t.date_string_col.right(2),
            lambda t: t.date_string_col.str[-2:],
            id='right',
        ),
        param(
            lambda t: t.date_string_col[1:3],
            lambda t: t.date_string_col.str[1:3],
            id='slice',
        ),
        param(
            lambda t: t.date_string_col[t.date_string_col.length() - 1 :],
            lambda t: t.date_string_col.str[-1:],
            id='expr_slice_begin',
        ),
        param(
            lambda t: t.date_string_col[: t.date_string_col.length()],
            lambda t: t.date_string_col,
            id='expr_slice_end',
        ),
        param(
            lambda t: t.date_string_col[:],
            lambda t: t.date_string_col,
            id='expr_empty_slice',
        ),
        param(
            lambda t: t.date_string_col[
                t.date_string_col.length() - 2 : t.date_string_col.length() - 1
            ],
            lambda t: t.date_string_col.str[-2:-1],
            id='expr_slice_begin_end',
        ),
        param(
            lambda t: t.date_string_col.split('/'),
            lambda t: t.date_string_col.str.split('/'),
            id='split',
        ),
        param(
            lambda t: ibis.literal('-').join(['a', t.string_col, 'c']),
            lambda t: 'a-' + t.string_col + '-c',
            id='join',
        ),
    ],
)
@pytest.mark.xfail_unsupported
def test_string(backend, alltypes, df, result_func, expected_func):
    expr = result_func(alltypes)
    result = expr.execute()

    expected = backend.default_series_rename(expected_func(df))
    backend.assert_series_equal(result, expected)


@pytest.mark.parametrize(
    'data, data_type',
    [param('123e4567-e89b-12d3-a456-426655440000', 'uuid', id='uuid')],
)
@pytest.mark.only_on_backends([PostgreSQL])
def test_special_strings(backend, con, alltypes, data, data_type):
    lit = ibis.literal(data, type=data_type).name('tmp')
    expr = alltypes[[alltypes.id, lit]].head(1)
    df = expr.execute()
    assert df['tmp'].iloc[0] == data



"""The pandas client implementation."""

from __future__ import absolute_import

import re
from functools import partial

import dateutil.parser
import numpy as np
import pandas as pd
import pytz
import toolz
from multipledispatch import Dispatcher
from pkg_resources import parse_version

import ibis.client as client
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.schema as sch
import ibis.expr.types as ir
from ibis.compat import CategoricalDtype, DatetimeTZDtype
from ibis.pandas.core import execute_and_reset

try:
    infer_pandas_dtype = pd.api.types.infer_dtype
except AttributeError:
    infer_pandas_dtype = pd.lib.infer_dtype


_ibis_dtypes = toolz.valmap(
    np.dtype,
    {
        dt.Boolean: np.bool_,
        dt.Null: np.object_,
        dt.Array: np.object_,
        dt.String: np.object_,
        dt.Binary: np.object_,
        dt.Date: 'datetime64[ns]',
        dt.Time: 'timedelta64[ns]',
        dt.Timestamp: 'datetime64[ns]',
        dt.Int8: np.int8,
        dt.Int16: np.int16,
        dt.Int32: np.int32,
        dt.Int64: np.int64,
        dt.UInt8: np.uint8,
        dt.UInt16: np.uint16,
        dt.UInt32: np.uint32,
        dt.UInt64: np.uint64,
        dt.Float32: np.float32,
        dt.Float64: np.float64,
        dt.Decimal: np.object_,
        dt.Struct: np.object_,
    },
)


_numpy_dtypes = toolz.keymap(
    np.dtype,
    {
        'bool': dt.boolean,
        'int8': dt.int8,
        'int16': dt.int16,
        'int32': dt.int32,
        'int64': dt.int64,
        'uint8': dt.uint8,
        'uint16': dt.uint16,
        'uint32': dt.uint32,
        'uint64': dt.uint64,
        'float16': dt.float16,
        'float32': dt.float32,
        'float64': dt.float64,
        'double': dt.double,
        'unicode': dt.string,
        'str': dt.string,
        'datetime64': dt.timestamp,
        'datetime64[ns]': dt.timestamp,
        'timedelta64': dt.interval,
        'timedelta64[ns]': dt.Interval('ns'),
    },
)


_inferable_pandas_dtypes = {
    'boolean': dt.boolean,
    'string': dt.string,
    'unicode': dt.string,
    'bytes': dt.string,
    'empty': dt.string,
}


@dt.dtype.register(np.dtype)
def from_numpy_dtype(value):
    try:
        return _numpy_dtypes[value]
    except KeyError:
        raise TypeError(
            'numpy dtype {!r} is not supported in the pandas backend'.format(
                value
            )
        )


@dt.dtype.register(DatetimeTZDtype)
def from_pandas_tzdtype(value):
    return dt.Timestamp(timezone=str(value.tz))


@dt.dtype.register(CategoricalDtype)
def from_pandas_categorical(value):
    return dt.Category()


@dt.infer.register(
    (np.generic,)
    + tuple(
        frozenset(
            np.signedinteger.__subclasses__()
            + np.unsignedinteger.__subclasses__()  # np.int64, np.uint64, etc.
        )
    )  # we need this because in Python 2 int is a parent of np.integer
)
def infer_numpy_scalar(value):
    return dt.dtype(value.dtype)


@dt.infer.register(pd.Timestamp)
def infer_pandas_timestamp(value):
    if value.tz is not None:
        return dt.Timestamp(timezone=str(value.tz))
    else:
        return dt.timestamp


@dt.infer.register(np.ndarray)
def infer_array(value):
    # TODO(kszucs): infer series
    return dt.Array(dt.dtype(value.dtype.name))


@sch.schema.register(pd.Series)
def schema_from_series(s):
    return sch.schema(tuple(s.iteritems()))


@sch.infer.register(pd.DataFrame)
def infer_pandas_schema(df, schema=None):
    schema = schema if schema is not None else {}

    pairs = []
    for column_name, pandas_dtype in df.dtypes.iteritems():
        if not isinstance(column_name, str):
            raise TypeError(
                'Column names must be strings to use the pandas backend'
            )

        if column_name in schema:
            ibis_dtype = dt.dtype(schema[column_name])
        elif pandas_dtype == np.object_:
            inferred_dtype = infer_pandas_dtype(df[column_name], skipna=True)
            if inferred_dtype in {'mixed', 'decimal'}:
                # TODO: in principal we can handle decimal (added in pandas
                # 0.23)
                raise TypeError(
                    'Unable to infer type of column {0!r}. Try instantiating '
                    'your table from the client with client.table('
                    "'my_table', schema={{{0!r}: <explicit type>}})".format(
                        column_name
                    )
                )
            ibis_dtype = _inferable_pandas_dtypes[inferred_dtype]
        else:
            ibis_dtype = dt.dtype(pandas_dtype)

        pairs.append((column_name, ibis_dtype))

    return sch.schema(pairs)


def ibis_dtype_to_pandas(ibis_dtype):
    """Convert ibis dtype to the pandas / numpy alternative"""
    assert isinstance(ibis_dtype, dt.DataType)

    if isinstance(ibis_dtype, dt.Timestamp) and ibis_dtype.timezone:
        return DatetimeTZDtype('ns', ibis_dtype.timezone)
    elif isinstance(ibis_dtype, dt.Interval):
        return np.dtype('timedelta64[{}]'.format(ibis_dtype.unit))
    elif isinstance(ibis_dtype, dt.Category):
        return CategoricalDtype()
    elif type(ibis_dtype) in _ibis_dtypes:
        return _ibis_dtypes[type(ibis_dtype)]
    else:
        return np.dtype(np.object_)


def ibis_schema_to_pandas(schema):
    return list(zip(schema.names, map(ibis_dtype_to_pandas, schema.types)))


convert = Dispatcher(
    'convert',
    doc="""\
Convert `column` to the pandas dtype corresponding to `out_dtype`, where the
dtype of `column` is `in_dtype`.

Parameters
----------
in_dtype : Union[np.dtype, pandas_dtype]
    The dtype of `column`, used for dispatching
out_dtype : ibis.expr.datatypes.DataType
    The requested ibis type of the output
column : pd.Series
    The column to convert

Returns
-------
result : pd.Series
    The converted column
""",
)


@convert.register(DatetimeTZDtype, dt.Timestamp, pd.Series)
def convert_datetimetz_to_timestamp(in_dtype, out_dtype, column):
    output_timezone = out_dtype.timezone
    if output_timezone is not None:
        return column.dt.tz_convert(output_timezone)
    return column.astype(out_dtype.to_pandas(), errors='ignore')


def convert_timezone(obj, timezone):
    """Convert `obj` to the timezone `timezone`.

    Parameters
    ----------
    obj : datetime.date or datetime.datetime

    Returns
    -------
    type(obj)
    """
    if timezone is None:
        return obj.replace(tzinfo=None)
    return pytz.timezone(timezone).localize(obj)


PANDAS_STRING_TYPES = {'string', 'unicode', 'bytes'}
PANDAS_DATE_TYPES = {'datetime', 'datetime64', 'date'}


@convert.register(np.dtype, dt.Timestamp, pd.Series)
def convert_datetime64_to_timestamp(in_dtype, out_dtype, column):
    if in_dtype.type == np.datetime64:
        return column.astype(out_dtype.to_pandas(), errors='ignore')
    try:
        series = pd.to_datetime(column, utc=True)
    except pd.errors.OutOfBoundsDatetime:
        inferred_dtype = infer_pandas_dtype(column, skipna=True)
        if inferred_dtype in PANDAS_DATE_TYPES:
            # not great, but not really any other option
            return column.map(
                partial(convert_timezone, timezone=out_dtype.timezone)
            )
        if inferred_dtype not in PANDAS_STRING_TYPES:
            raise TypeError(
                (
                    'Conversion to timestamp not supported for Series of type '
                    '{!r}'
                ).format(inferred_dtype)
            )
        return column.map(dateutil.parser.parse)
    else:
        utc_dtype = DatetimeTZDtype('ns', 'UTC')
        return series.astype(utc_dtype).dt.tz_convert(out_dtype.timezone)


@convert.register(np.dtype, dt.Interval, pd.Series)
def convert_any_to_interval(_, out_dtype, column):
    return column.values.astype(out_dtype.to_pandas())


@convert.register(np.dtype, dt.String, pd.Series)
def convert_any_to_string(_, out_dtype, column):
    result = column.astype(out_dtype.to_pandas(), errors='ignore')
    return result


@convert.register(np.dtype, dt.Boolean, pd.Series)
def convert_boolean_to_series(in_dtype, out_dtype, column):
    # XXX: this is a workaround until #1595 can be addressed
    in_dtype_type = in_dtype.type
    out_dtype_type = out_dtype.to_pandas().type
    if in_dtype_type != np.object_ and in_dtype_type != out_dtype_type:
        return column.astype(out_dtype_type)
    return column


@convert.register(object, dt.DataType, pd.Series)
def convert_any_to_any(_, out_dtype, column):
    return column.astype(out_dtype.to_pandas(), errors='ignore')


def ibis_schema_apply_to(schema, df):
    """Applies the Ibis schema to a pandas DataFrame

    Parameters
    ----------
    schema : ibis.schema.Schema
    df : pandas.DataFrame

    Returns
    -------
    df : pandas.DataFrame

    Notes
    -----
    Mutates `df`
    """

    for column, dtype in schema.items():
        pandas_dtype = dtype.to_pandas()
        col = df[column]
        col_dtype = col.dtype

        try:
            not_equal = pandas_dtype != col_dtype
        except TypeError:
            # ugh, we can't compare dtypes coming from pandas, assume not equal
            not_equal = True

        if not_equal or isinstance(dtype, dt.String):
            df[column] = convert(col_dtype, dtype, col)

    return df


dt.DataType.to_pandas = ibis_dtype_to_pandas
sch.Schema.to_pandas = ibis_schema_to_pandas
sch.Schema.apply_to = ibis_schema_apply_to


class PandasTable(ops.DatabaseTable):
    pass


class PandasClient(client.Client):

    dialect = None  # defined in ibis.pandas.api

    def __init__(self, dictionary):
        self.dictionary = dictionary

    def table(self, name, schema=None):
        df = self.dictionary[name]
        schema = sch.infer(df, schema=schema)
        return PandasTable(name, schema, self).to_expr()

    def execute(self, query, params=None, limit='default', **kwargs):
        if limit != 'default':
            raise ValueError(
                'limit parameter to execute is not yet implemented in the '
                'pandas backend'
            )

        if not isinstance(query, ir.Expr):
            raise TypeError(
                "`query` has type {!r}, expected ibis.expr.types.Expr".format(
                    type(query).__name__
                )
            )
        return execute_and_reset(query, params=params, **kwargs)

    def compile(self, expr, *args, **kwargs):
        """Compile `expr`.

        Notes
        -----
        For the pandas backend this is a no-op.

        """
        return expr

    def database(self, name=None):
        """Construct a database called `name`."""
        return PandasDatabase(name, self)

    def list_tables(self, like=None):
        """List the available tables."""
        tables = list(self.dictionary.keys())
        if like is not None:
            pattern = re.compile(like)
            return list(filter(lambda t: pattern.findall(t), tables))
        return tables

    def load_data(self, table_name, obj, **kwargs):
        """Load data from `obj` into `table_name`.

        Parameters
        ----------
        table_name : str
        obj : pandas.DataFrame

        """
        # kwargs is a catch all for any options required by other backends.
        self.dictionary[table_name] = pd.DataFrame(obj)

    def create_table(self, table_name, obj=None, schema=None):
        """Create a table."""
        if obj is None and schema is None:
            raise com.IbisError('Must pass expr or schema')

        if obj is not None:
            df = pd.DataFrame(obj)
        else:
            dtypes = ibis_schema_to_pandas(schema)
            df = schema.apply_to(
                pd.DataFrame(columns=list(map(toolz.first, dtypes)))
            )

        self.dictionary[table_name] = df

    def get_schema(self, table_name, database=None):
        """Return a Schema object for the indicated table and database.

        Parameters
        ----------
        table_name : str
            May be fully qualified
        database : str

        Returns
        -------
        ibis.expr.schema.Schema

        """
        return sch.infer(self.dictionary[table_name])

    def exists_table(self, name):
        """Determine if the indicated table or view exists.

        Parameters
        ----------
        name : str
        database : str

        Returns
        -------
        bool

        """
        return bool(self.list_tables(like=name))

    @property
    def version(self) -> str:
        """Return the version of the underlying backend library."""
        return parse_version(pd.__version__)


class PandasDatabase(client.Database):
    pass



import itertools
import os
import webbrowser

import numpy as np

import ibis
import ibis.common.exceptions as com
import ibis.config as config
import ibis.util as util

# TODO move methods containing ops import to api.py


class Expr:
    """Base expression class"""

    def _type_display(self):
        return type(self).__name__

    def __init__(self, arg):
        # TODO: all inputs must inherit from a common table API
        self._arg = arg

    def __repr__(self):
        if not config.options.interactive:
            return self._repr()

        try:
            result = self.execute()
        except com.TranslationError as e:
            output = (
                'Translation to backend failed\n'
                'Error message: {0}\n'
                'Expression repr follows:\n{1}'.format(e.args[0], self._repr())
            )
            return output
        else:
            return repr(result)

    def __hash__(self):
        return hash(self._key)

    def __bool__(self):
        raise ValueError(
            "The truth value of an Ibis expression is not " "defined"
        )

    __nonzero__ = __bool__

    def _repr(self, memo=None):
        from ibis.expr.format import ExprFormatter

        return ExprFormatter(self, memo=memo).get_result()

    @property
    def _safe_name(self):
        """Get the name of an expression `expr`, returning ``None`` if the
        expression has no name.

        Returns
        -------
        Optional[str]
        """
        try:
            return self.get_name()
        except (com.ExpressionError, AttributeError):
            return None

    @property
    def _key(self):
        """Key suitable for hashing an expression.

        Returns
        -------
        Tuple[Type[Expr], Optional[str], ibis.expr.operations.Node]
            A tuple of hashable objects uniquely identifying this expression.
        """
        return type(self), self._safe_name, self.op()

    def _repr_png_(self):
        if not ibis.options.graphviz_repr:
            return None
        try:
            import ibis.expr.visualize as viz
        except ImportError:
            return None
        else:
            try:
                return viz.to_graph(self).pipe(format='png')
            except Exception:
                # Something may go wrong, and we can't error in the notebook
                # so fallback to the default text representation.
                return None

    def visualize(self, format='svg'):
        """Visualize an expression in the browser as an SVG image.

        Parameters
        ----------
        format : str, optional
            Defaults to ``'svg'``. Some additional formats are
            ``'jpeg'`` and ``'png'``. These are specified by the ``graphviz``
            Python library.

        Notes
        -----
        This method opens a web browser tab showing the image of the expression
        graph created by the code in :module:`ibis.expr.visualize`.

        Raises
        ------
        ImportError
            If ``graphviz`` is not installed.
        """
        import ibis.expr.visualize as viz

        path = viz.draw(viz.to_graph(self), format=format)
        webbrowser.open('file://{}'.format(os.path.abspath(path)))

    def pipe(self, f, *args, **kwargs):
        """Generic composition function to enable expression pipelining.

        Parameters
        ----------
        f : function or (function, arg_name) tuple
          If the expression needs to be passed as anything other than the first
          argument to the function, pass a tuple with the argument name. For
          example, (f, 'data') if the function f expects a 'data' keyword
        args : positional arguments
        kwargs : keyword arguments

        Examples
        --------
        >>> import ibis
        >>> t = ibis.table([('a', 'int64'), ('b', 'string')], name='t')
        >>> f = lambda a: (a + 1).name('a')
        >>> g = lambda a: (a * 2).name('a')
        >>> result1 = t.a.pipe(f).pipe(g)
        >>> result1  # doctest: +NORMALIZE_WHITESPACE
        ref_0
        UnboundTable[table]
          name: t
          schema:
            a : int64
            b : string
        a = Multiply[int64*]
          left:
            a = Add[int64*]
              left:
                a = Column[int64*] 'a' from table
                  ref_0
              right:
                Literal[int8]
                  1
          right:
            Literal[int8]
              2
        >>> result2 = g(f(t.a))  # equivalent to the above
        >>> result1.equals(result2)
        True

        Returns
        -------
        result : result type of passed function
        """
        if isinstance(f, tuple):
            f, data_keyword = f
            kwargs = kwargs.copy()
            kwargs[data_keyword] = self
            return f(*args, **kwargs)
        else:
            return f(self, *args, **kwargs)

    __call__ = pipe

    def op(self):
        return self._arg

    @property
    def _factory(self):
        return type(self)

    def execute(self, limit='default', params=None, **kwargs):
        """
        If this expression is based on physical tables in a database backend,
        execute it against that backend.

        Parameters
        ----------
        limit : integer or None, default 'default'
          Pass an integer to effect a specific row limit. limit=None means "no
          limit". The default is whatever is in ibis.options.

        Returns
        -------
        result : expression-dependent
          Result of compiling expression and executing in backend
        """
        from ibis.client import execute

        return execute(self, limit=limit, params=params, **kwargs)

    def compile(self, limit=None, params=None):
        """
        Compile expression to whatever execution target, to verify

        Returns
        -------
        compiled : value or list
           query representation or list thereof
        """
        from ibis.client import compile

        return compile(self, limit=limit, params=params)

    def verify(self):
        """
        Returns True if expression can be compiled to its attached client
        """
        try:
            self.compile()
        except Exception:
            return False
        else:
            return True

    def equals(self, other, cache=None):
        if type(self) != type(other):
            return False
        return self._arg.equals(other._arg, cache=cache)

    def _root_tables(self):
        return self.op().root_tables()


class ExprList(Expr):
    def _type_display(self):
        return ', '.join(expr._type_display() for expr in self.exprs())

    def exprs(self):
        return self.op().exprs

    def names(self):
        return [x.get_name() for x in self.exprs()]

    def types(self):
        return [x.type() for x in self.exprs()]

    def schema(self):
        import ibis.expr.schema as sch

        return sch.Schema(self.names(), self.types())

    def rename(self, f):
        import ibis.expr.operations as ops

        new_exprs = [x.name(f(x.get_name())) for x in self.exprs()]
        return ops.ExpressionList(new_exprs).to_expr()

    def prefix(self, value):
        return self.rename(lambda x: value + x)

    def suffix(self, value):
        return self.rename(lambda x: x + value)

    def concat(self, *others):
        """
        Concatenate expression lists

        Returns
        -------
        combined : ExprList
        """
        import ibis.expr.operations as ops

        exprs = list(self.exprs())
        for o in others:
            if not isinstance(o, ExprList):
                raise TypeError(o)
            exprs.extend(o.exprs())
        return ops.ExpressionList(exprs).to_expr()


# ---------------------------------------------------------------------
# Helper / factory functions


class ValueExpr(Expr):

    """
    Base class for a data generating expression having a fixed and known type,
    either a single value (scalar)
    """

    def __init__(self, arg, dtype, name=None):
        super().__init__(arg)
        self._name = name
        self._dtype = dtype

    def equals(self, other, cache=None):
        return (
            isinstance(other, ValueExpr)
            and self._name == other._name
            and self._dtype == other._dtype
            and super().equals(other, cache=cache)
        )

    def has_name(self):
        if self._name is not None:
            return True
        return self.op().has_resolved_name()

    def get_name(self):
        if self._name is not None:
            # This value has been explicitly named
            return self._name

        # In some but not all cases we can get a name from the node that
        # produces the value
        return self.op().resolve_name()

    def name(self, name):
        return self._factory(self._arg, name=name)

    def type(self):
        return self._dtype

    @property
    def _factory(self):
        def factory(arg, name=None):
            return type(self)(arg, dtype=self.type(), name=name)

        return factory


class ScalarExpr(ValueExpr):
    def _type_display(self):
        return str(self.type())


class ColumnExpr(ValueExpr):
    def _type_display(self):
        return '{}*'.format(self.type())

    def parent(self):
        return self._arg

    def to_projection(self):
        """
        Promote this column expression to a table projection
        """
        roots = self._root_tables()
        if len(roots) > 1:
            raise com.RelationError(
                'Cannot convert array expression '
                'involving multiple base table references '
                'to a projection'
            )

        table = TableExpr(roots[0])
        return table.projection([self])


class AnalyticExpr(Expr):
    @property
    def _factory(self):
        def factory(arg):
            return type(self)(arg)

        return factory

    def _type_display(self):
        return str(self.type())

    def type(self):
        return 'analytic'


class TableExpr(Expr):
    @property
    def _factory(self):
        def factory(arg):
            return TableExpr(arg)

        return factory

    def _type_display(self):
        return 'table'

    def _is_valid(self, exprs):
        try:
            self._assert_valid(util.promote_list(exprs))
        except com.RelationError:
            return False
        else:
            return True

    def _assert_valid(self, exprs):
        from ibis.expr.analysis import ExprValidator

        ExprValidator([self]).validate_all(exprs)

    def __contains__(self, name):
        return name in self.schema()

    def __getitem__(self, what):
        if isinstance(what, (str, int)):
            return self.get_column(what)

        if isinstance(what, slice):
            step = what.step
            if step is not None and step != 1:
                raise ValueError('Slice step can only be 1')
            start = what.start or 0
            stop = what.stop

            if stop is None or stop < 0:
                raise ValueError('End index must be a positive number')

            if start < 0:
                raise ValueError('Start index must be a positive number')

            return self.limit(stop - start, offset=start)

        what = bind_expr(self, what)

        if isinstance(what, AnalyticExpr):
            what = what._table_getitem()

        if isinstance(what, (list, tuple, TableExpr)):
            # Projection case
            return self.projection(what)
        elif isinstance(what, BooleanColumn):
            # Boolean predicate
            return self.filter([what])
        elif isinstance(what, ColumnExpr):
            # Projection convenience
            return self.projection(what)
        else:
            raise NotImplementedError(
                'Selection rows or columns with {} objects is not '
                'supported'.format(type(what).__name__)
            )

    def __len__(self):
        raise com.ExpressionError('Use .count() instead')

    def __setstate__(self, instance_dictionary):
        self.__dict__ = instance_dictionary

    def __getattr__(self, key):
        try:
            schema = self.schema()
        except com.IbisError:
            raise AttributeError(key)

        if key not in schema:
            raise AttributeError(key)

        try:
            return self.get_column(key)
        except com.IbisTypeError:
            raise AttributeError(key)

    def __dir__(self):
        attrs = dir(type(self))
        if self._is_materialized():
            attrs = frozenset(attrs + self.schema().names)
        return sorted(attrs)

    def _resolve(self, exprs):
        exprs = util.promote_list(exprs)
        return list(map(self._ensure_expr, exprs))

    def _ensure_expr(self, expr):
        if isinstance(expr, str):
            return self[expr]
        elif isinstance(expr, (int, np.integer)):
            return self[self.schema().name_at_position(expr)]
        elif not isinstance(expr, Expr):
            return expr(self)
        else:
            return expr

    def _get_type(self, name):
        return self._arg.get_type(name)

    def get_columns(self, iterable):
        """
        Get multiple columns from the table

        Examples
        --------
        >>> import ibis
        >>> table = ibis.table(
        ...    [
        ...        ('a', 'int64'),
        ...        ('b', 'string'),
        ...        ('c', 'timestamp'),
        ...        ('d', 'float'),
        ...    ],
        ...    name='t'
        ... )
        >>> a, b, c = table.get_columns(['a', 'b', 'c'])

        Returns
        -------
        columns : list of column/array expressions
        """
        return [self.get_column(x) for x in iterable]

    def get_column(self, name):
        """
        Get a reference to a single column from the table

        Returns
        -------
        column : array expression
        """
        import ibis.expr.operations as ops

        ref = ops.TableColumn(name, self)
        return ref.to_expr()

    @property
    def columns(self):
        return self.schema().names

    def schema(self):
        """
        Get the schema for this table (if one is known)

        Returns
        -------
        schema : Schema
        """
        if not self._is_materialized():
            raise com.IbisError('Table operation is not yet materialized')
        return self.op().schema

    def _is_materialized(self):
        # The operation produces a known schema
        return self.op().has_schema()

    def group_by(self, by=None, **additional_grouping_expressions):
        """
        Create an intermediate grouped table expression, pending some group
        operation to be applied with it.

        Examples
        --------
        >>> import ibis
        >>> pairs = [('a', 'int32'), ('b', 'timestamp'), ('c', 'double')]
        >>> t = ibis.table(pairs)
        >>> b1, b2 = t.a, t.b
        >>> result = t.group_by([b1, b2]).aggregate(sum_of_c=t.c.sum())

        Notes
        -----
        group_by and groupby are equivalent, with `groupby` being provided for
        ease-of-use for pandas users.

        Returns
        -------
        grouped_expr : GroupedTableExpr
        """
        from ibis.expr.groupby import GroupedTableExpr

        return GroupedTableExpr(self, by, **additional_grouping_expressions)

    groupby = group_by


# -----------------------------------------------------------------------------
# Declare all typed ValueExprs. This is what the user will actually interact
# with: an instance of each is well-typed and includes all valid methods
# defined for each type.


class AnyValue(ValueExpr):
    pass  # noqa: E701,E302


class AnyScalar(ScalarExpr, AnyValue):
    pass  # noqa: E701,E302


class AnyColumn(ColumnExpr, AnyValue):
    pass  # noqa: E701,E302


class NullValue(AnyValue):
    pass  # noqa: E701,E302


class NullScalar(AnyScalar, NullValue):
    pass  # noqa: E701,E302


class NullColumn(AnyColumn, NullValue):
    pass  # noqa: E701,E302


class NumericValue(AnyValue):
    pass  # noqa: E701,E302


class NumericScalar(AnyScalar, NumericValue):
    pass  # noqa: E701,E302


class NumericColumn(AnyColumn, NumericValue):
    pass  # noqa: E701,E302


class BooleanValue(NumericValue):
    pass  # noqa: E701,E302


class BooleanScalar(NumericScalar, BooleanValue):
    pass  # noqa: E701,E302


class BooleanColumn(NumericColumn, BooleanValue):
    pass  # noqa: E701,E302


class IntegerValue(NumericValue):
    pass  # noqa: E701,E302


class IntegerScalar(NumericScalar, IntegerValue):
    pass  # noqa: E701,E302


class IntegerColumn(NumericColumn, IntegerValue):
    pass  # noqa: E701,E302


class FloatingValue(NumericValue):
    pass  # noqa: E701,E302


class FloatingScalar(NumericScalar, FloatingValue):
    pass  # noqa: E701,E302


class FloatingColumn(NumericColumn, FloatingValue):
    pass  # noqa: E701,E302


class DecimalValue(NumericValue):
    pass  # noqa: E701,E302


class DecimalScalar(NumericScalar, DecimalValue):
    pass  # noqa: E701,E302


class DecimalColumn(NumericColumn, DecimalValue):
    pass  # noqa: E701,E302


class StringValue(AnyValue):
    pass  # noqa: E701,E302


class StringScalar(AnyScalar, StringValue):
    pass  # noqa: E701,E302


class StringColumn(AnyColumn, StringValue):
    pass  # noqa: E701,E302


class BinaryValue(AnyValue):
    pass  # noqa: E701,E302


class BinaryScalar(AnyScalar, BinaryValue):
    pass  # noqa: E701,E302


class BinaryColumn(AnyColumn, BinaryValue):
    pass  # noqa: E701,E302


class TemporalValue(AnyValue):
    pass  # noqa: E701,E302


class TemporalScalar(AnyScalar, TemporalValue):
    pass  # noqa: E701,E302


class TemporalColumn(AnyColumn, TemporalValue):
    pass  # noqa: E701,E302


class TimeValue(TemporalValue):
    pass  # noqa: E701,E302


class TimeScalar(TemporalScalar, TimeValue):
    pass  # noqa: E701,E302


class TimeColumn(TemporalColumn, TimeValue):
    pass  # noqa: E701,E302


class DateValue(TemporalValue):
    pass  # noqa: E701,E302


class DateScalar(TemporalScalar, DateValue):
    pass  # noqa: E701,E302


class DateColumn(TemporalColumn, DateValue):
    pass  # noqa: E701,E302


class TimestampValue(TemporalValue):
    pass  # noqa: E701,E302


class TimestampScalar(TemporalScalar, TimestampValue):
    pass  # noqa: E701,E302


class TimestampColumn(TemporalColumn, TimestampValue):
    pass  # noqa: E701,E302


class CategoryValue(AnyValue):
    pass  # noqa: E701,E302


class CategoryScalar(AnyScalar, CategoryValue):
    pass  # noqa: E701,E302


class CategoryColumn(AnyColumn, CategoryValue):
    pass  # noqa: E701,E302


class EnumValue(AnyValue):
    pass  # noqa: E701,E302


class EnumScalar(AnyScalar, EnumValue):
    pass  # noqa: E701,E302


class EnumColumn(AnyColumn, EnumValue):
    pass  # noqa: E701,E302


class ArrayValue(AnyValue):
    pass  # noqa: E701,E302


class ArrayScalar(AnyScalar, ArrayValue):
    pass  # noqa: E701,E302


class ArrayColumn(AnyColumn, ArrayValue):
    pass  # noqa: E701,E302


class SetValue(AnyValue):
    pass  # noqa: E701,E302


class SetScalar(AnyScalar, SetValue):
    pass  # noqa: E701,E302


class SetColumn(AnyColumn, SetValue):
    pass  # noqa: E701,E302


class MapValue(AnyValue):
    pass  # noqa: E701,E302


class MapScalar(AnyScalar, MapValue):
    pass  # noqa: E701,E302


class MapColumn(AnyColumn, MapValue):
    pass  # noqa: E701,E302


class JSONValue(StringValue):
    pass  # noqa: E701,E302


class JSONScalar(StringScalar, JSONValue):
    pass  # noqa: E701,E302


class JSONColumn(StringColumn, JSONValue):
    pass  # noqa: E701,E302


class JSONBValue(BinaryValue):
    pass  # noqa: E701,E302


class JSONBScalar(BinaryScalar, JSONBValue):
    pass  # noqa: E701,E302


class JSONBColumn(BinaryColumn, JSONBValue):
    pass  # noqa: E701,E302


class StructValue(AnyValue):
    def __dir__(self):
        return sorted(
            frozenset(itertools.chain(dir(type(self)), self.type().names))
        )


class StructScalar(AnyScalar, StructValue):
    pass  # noqa: E701,E302


class StructColumn(AnyColumn, StructValue):
    pass  # noqa: E701,E302


class IntervalValue(AnyValue):
    pass  # noqa: E701,E302


class IntervalScalar(AnyScalar, IntervalValue):
    pass  # noqa: E701,E302


class IntervalColumn(AnyColumn, IntervalValue):
    pass  # noqa: E701,E302


class GeoSpatialValue(NumericValue):
    pass  # noqa: E701,E302


class GeoSpatialScalar(NumericScalar, GeoSpatialValue):
    pass  # noqa: E701,E302,E501


class GeoSpatialColumn(NumericColumn, GeoSpatialValue):
    pass  # noqa: E701,E302,E501


class PointValue(GeoSpatialValue):
    pass  # noqa: E701,E302


class PointScalar(GeoSpatialScalar, PointValue):
    pass  # noqa: E701,E302


class PointColumn(GeoSpatialColumn, PointValue):
    pass  # noqa: E701,E302


class LineStringValue(GeoSpatialValue):
    pass  # noqa: E701,E302


class LineStringScalar(GeoSpatialScalar, LineStringValue):
    pass  # noqa: E701,E302,E501


class LineStringColumn(GeoSpatialColumn, LineStringValue):
    pass  # noqa: E701,E302,E501


class PolygonValue(GeoSpatialValue):
    pass  # noqa: E701,E302


class PolygonScalar(GeoSpatialScalar, PolygonValue):
    pass  # noqa: E701,E302


class PolygonColumn(GeoSpatialColumn, PolygonValue):
    pass  # noqa: E701,E302


class MultiLineStringValue(GeoSpatialValue):
    pass  # noqa: E701,E302


class MultiLineStringScalar(
    GeoSpatialScalar, MultiLineStringValue
):  # noqa: E302
    pass  # noqa: E701


class MultiLineStringColumn(
    GeoSpatialColumn, MultiLineStringValue
):  # noqa: E302
    pass  # noqa: E701


class MultiPointValue(GeoSpatialValue):
    pass  # noqa: E701,E302


class MultiPointScalar(GeoSpatialScalar, MultiPointValue):  # noqa: E302
    pass  # noqa: E701


class MultiPointColumn(GeoSpatialColumn, MultiPointValue):  # noqa: E302
    pass  # noqa: E701


class MultiPolygonValue(GeoSpatialValue):
    pass  # noqa: E701,E302


class MultiPolygonScalar(GeoSpatialScalar, MultiPolygonValue):  # noqa: E302
    pass  # noqa: E701


class MultiPolygonColumn(GeoSpatialColumn, MultiPolygonValue):  # noqa: E302
    pass  # noqa: E701


class UUIDValue(StringValue):
    pass  # noqa: E701,E302


class UUIDScalar(StringScalar, UUIDValue):
    pass  # noqa: E701,E302


class UUIDColumn(StringColumn, UUIDValue):
    pass  # noqa: E701,E302


class ListExpr(ColumnExpr, AnyValue):
    @property
    def values(self):
        return self.op().values

    def __iter__(self):
        return iter(self.values)

    def __getitem__(self, key):
        return self.values[key]

    def __add__(self, other):
        other_values = tuple(getattr(other, 'values', other))
        return type(self.op())(self.values + other_values).to_expr()

    def __radd__(self, other):
        other_values = tuple(getattr(other, 'values', other))
        return type(self.op())(other_values + self.values).to_expr()

    def __bool__(self):
        return bool(self.values)

    __nonzero__ = __bool__

    def __len__(self):
        return len(self.values)


class TopKExpr(AnalyticExpr):
    def type(self):
        return 'topk'

    def _table_getitem(self):
        return self.to_filter()

    def to_filter(self):
        # TODO: move to api.py
        import ibis.expr.operations as ops

        return ops.SummaryFilter(self).to_expr()

    def to_aggregation(
        self, metric_name=None, parent_table=None, backup_metric_name=None
    ):
        """
        Convert the TopK operation to a table aggregation
        """
        op = self.op()

        arg_table = find_base_table(op.arg)

        by = op.by
        if not isinstance(by, Expr):
            by = by(arg_table)
            by_table = arg_table
        else:
            by_table = find_base_table(op.by)

        if metric_name is None:
            if by.get_name() == op.arg.get_name():
                by = by.name(backup_metric_name)
        else:
            by = by.name(metric_name)

        if arg_table.equals(by_table):
            agg = arg_table.aggregate(by, by=[op.arg])
        elif parent_table is not None:
            agg = parent_table.aggregate(by, by=[op.arg])
        else:
            raise com.IbisError(
                'Cross-table TopK; must provide a parent ' 'joined table'
            )

        return agg.sort_by([(by.get_name(), False)]).limit(op.k)


class SortExpr(Expr):
    def _type_display(self):
        return 'array-sort'

    def get_name(self):
        return self.op().resolve_name()


class DayOfWeek(Expr):
    def index(self):
        """Get the index of the day of the week.

        Returns
        -------
        IntegerValue
            The index of the day of the week. Ibis follows pandas conventions,
            where **Monday = 0 and Sunday = 6**.
        """
        import ibis.expr.operations as ops

        return ops.DayOfWeekIndex(self.op().arg).to_expr()

    def full_name(self):
        """Get the name of the day of the week.

        Returns
        -------
        StringValue
            The name of the day of the week
        """
        import ibis.expr.operations as ops

        return ops.DayOfWeekName(self.op().arg).to_expr()


def bind_expr(table, expr):
    if isinstance(expr, (list, tuple)):
        return [bind_expr(table, x) for x in expr]

    return table._ensure_expr(expr)


# TODO: move to analysis
def find_base_table(expr):
    if isinstance(expr, TableExpr):
        return expr

    for arg in expr.op().flat_args():
        if isinstance(arg, Expr):
            r = find_base_table(arg)
            if isinstance(r, TableExpr):
                return r


_NULL = None


def null():
    """Create a NULL/NA scalar"""
    import ibis.expr.operations as ops

    global _NULL
    if _NULL is None:
        _NULL = ops.NullLiteral().to_expr()

    return _NULL


def literal(value, type=None):
    """Create a scalar expression from a Python value.

    Parameters
    ----------
    value : some Python basic type
        A Python value
    type : ibis type or string, optional
        An instance of :class:`ibis.expr.datatypes.DataType` or a string
        indicating the ibis type of `value`. This parameter should only be used
        in cases where ibis's type inference isn't sufficient for discovering
        the type of `value`.

    Returns
    -------
    literal_value : Literal
        An expression representing a literal value

    Examples
    --------
    >>> import ibis
    >>> x = ibis.literal(42)
    >>> x.type()
    int8
    >>> y = ibis.literal(42, type='double')
    >>> y.type()
    float64
    >>> ibis.literal('foobar', type='int64')  # doctest: +ELLIPSIS
    Traceback (most recent call last):
      ...
    TypeError: Value 'foobar' cannot be safely coerced to int64
    """
    import ibis.expr.datatypes as dt
    import ibis.expr.operations as ops

    if hasattr(value, 'op') and isinstance(value.op(), ops.Literal):
        return value

    try:
        inferred_dtype = dt.infer(value)
    except com.InputTypeError:
        has_inferred = False
    else:
        has_inferred = True

    if type is None:
        has_explicit = False
    else:
        has_explicit = True
        explicit_dtype = dt.dtype(type)

    if has_explicit and has_inferred:
        try:
            # ensure type correctness: check that the inferred dtype is
            # implicitly castable to the explicitly given dtype and value
            dtype = inferred_dtype.cast(explicit_dtype, value=value)
        except com.IbisTypeError:
            raise TypeError(
                'Value {!r} cannot be safely coerced to {}'.format(value, type)
            )
    elif has_explicit:
        dtype = explicit_dtype
    elif has_inferred:
        dtype = inferred_dtype
    else:
        raise TypeError(
            'The datatype of value {!r} cannot be inferred, try '
            'passing it explicitly with the `type` keyword.'.format(value)
        )

    if dtype is dt.null:
        return null().cast(dtype)
    else:
        return ops.Literal(value, dtype=dtype).to_expr()


def sequence(values):
    """
    Wrap a list of Python values as an Ibis sequence type

    Parameters
    ----------
    values : list
      Should all be None or the same type

    Returns
    -------
    seq : Sequence
    """
    import ibis.expr.operations as ops

    return ops.ValueList(values).to_expr()


def as_value_expr(val):
    import pandas as pd

    if not isinstance(val, Expr):
        if isinstance(val, (tuple, list)):
            val = sequence(val)
        elif isinstance(val, pd.Series):
            val = sequence(list(val))
        else:
            val = literal(val)

    return val


def param(type):
    """Create a parameter of a particular type to be defined just before
    execution.

    Parameters
    ----------
    type : dt.DataType
        The type of the unbound parameter, e.g., double, int64, date, etc.

    Returns
    -------
    ScalarExpr

    Examples
    --------
    >>> import ibis
    >>> import ibis.expr.datatypes as dt
    >>> start = ibis.param(dt.date)
    >>> end = ibis.param(dt.date)
    >>> schema = [('timestamp_col', 'timestamp'), ('value', 'double')]
    >>> t = ibis.table(schema)
    >>> predicates = [t.timestamp_col >= start, t.timestamp_col <= end]
    >>> expr = t.filter(predicates).value.sum()
    """
    import ibis.expr.datatypes as dt
    import ibis.expr.operations as ops

    return ops.ScalarParameter(dt.dtype(type)).to_expr()


class UnnamedMarker:
    pass


unnamed = UnnamedMarker()



import datetime

import numpy as np
import pandas as pd
from pandas.core.groupby import SeriesGroupBy

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
from ibis.pandas.core import (
    date_types,
    integer_types,
    numeric_types,
    timedelta_types,
    timestamp_types,
)
from ibis.pandas.dispatch import execute_node, pre_execute


@execute_node.register(ops.Strftime, pd.Timestamp, str)
def execute_strftime_timestamp_str(op, data, format_string, **kwargs):
    return data.strftime(format_string)


@execute_node.register(ops.Strftime, pd.Series, str)
def execute_strftime_series_str(op, data, format_string, **kwargs):
    return data.dt.strftime(format_string)


@execute_node.register(ops.ExtractTemporalField, pd.Timestamp)
def execute_extract_timestamp_field_timestamp(op, data, **kwargs):
    field_name = type(op).__name__.lower().replace('extract', '')
    return getattr(data, field_name)


@execute_node.register(ops.ExtractMillisecond, pd.Timestamp)
def execute_extract_millisecond_timestamp(op, data, **kwargs):
    return int(data.microsecond // 1000.0)


@execute_node.register(ops.ExtractTemporalField, pd.Series)
def execute_extract_timestamp_field_series(op, data, **kwargs):
    field_name = type(op).__name__.lower().replace('extract', '')
    return getattr(data.dt, field_name).astype(np.int32)


@execute_node.register(
    ops.BetweenTime,
    pd.Series,
    (pd.Series, str, datetime.time),
    (pd.Series, str, datetime.time),
)
def execute_between_time(op, data, lower, upper, **kwargs):
    indexer = pd.DatetimeIndex(data).indexer_between_time(lower, upper)
    result = np.zeros(len(data), dtype=np.bool_)
    result[indexer] = True
    return pd.Series(result)


@execute_node.register(ops.Date, pd.Series)
def execute_timestamp_date(op, data, **kwargs):
    return data.dt.floor('d')


@execute_node.register((ops.TimestampTruncate, ops.DateTruncate), pd.Series)
def execute_timestamp_truncate(op, data, **kwargs):
    dtype = 'datetime64[{}]'.format(op.unit)
    array = data.values.astype(dtype)
    return pd.Series(array, name=data.name)


OFFSET_CLASS = {
    "Y": pd.offsets.DateOffset,
    "Q": pd.offsets.DateOffset,
    "M": pd.offsets.DateOffset,
    "W": pd.offsets.DateOffset,
    # all other units are timedelta64s
}


@execute_node.register(ops.IntervalFromInteger, pd.Series)
def execute_interval_from_integer_series(op, data, **kwargs):
    unit = op.unit
    resolution = "{}s".format(op.resolution)
    cls = OFFSET_CLASS.get(unit, None)

    # fast path for timedelta conversion
    if cls is None:
        return data.astype("timedelta64[{}]".format(unit))
    return data.apply(
        lambda n, cls=cls, resolution=resolution: cls(**{resolution: n})
    )


@execute_node.register(ops.IntervalFromInteger, integer_types)
def execute_interval_from_integer_integer_types(op, data, **kwargs):
    unit = op.unit
    resolution = "{}s".format(op.resolution)
    cls = OFFSET_CLASS.get(unit, None)

    if cls is None:
        return pd.Timedelta(data, unit=unit)
    return cls(**{resolution: data})


@execute_node.register(ops.Cast, pd.Series, dt.Interval)
def execute_cast_integer_to_interval_series(op, data, type, **kwargs):
    to = op.to
    unit = to.unit
    resolution = "{}s".format(to.resolution)
    cls = OFFSET_CLASS.get(unit, None)

    if cls is None:
        return data.astype("timedelta64[{}]".format(unit))
    return data.apply(
        lambda n, cls=cls, resolution=resolution: cls(**{resolution: n})
    )


@execute_node.register(ops.Cast, integer_types, dt.Interval)
def execute_cast_integer_to_interval_integer_types(op, data, type, **kwargs):
    to = op.to
    unit = to.unit
    resolution = "{}s".format(to.resolution)
    cls = OFFSET_CLASS.get(unit, None)

    if cls is None:
        return pd.Timedelta(data, unit=unit)
    return cls(**{resolution: data})


@execute_node.register(ops.TimestampAdd, timestamp_types, timedelta_types)
def execute_timestamp_add_datetime_timedelta(op, left, right, **kwargs):
    return pd.Timestamp(left) + pd.Timedelta(right)


@execute_node.register(ops.TimestampAdd, timestamp_types, pd.Series)
def execute_timestamp_add_datetime_series(op, left, right, **kwargs):
    return pd.Timestamp(left) + right


@execute_node.register(ops.IntervalAdd, timedelta_types, timedelta_types)
def execute_interval_add_delta_delta(op, left, right, **kwargs):
    return op.op(pd.Timedelta(left), pd.Timedelta(right))


@execute_node.register(ops.IntervalAdd, timedelta_types, pd.Series)
@execute_node.register(
    ops.IntervalMultiply, timedelta_types, numeric_types + (pd.Series,)
)
def execute_interval_add_multiply_delta_series(op, left, right, **kwargs):
    return op.op(pd.Timedelta(left), right)


@execute_node.register(
    (ops.TimestampAdd, ops.IntervalAdd), pd.Series, timedelta_types
)
def execute_timestamp_interval_add_series_delta(op, left, right, **kwargs):
    return left + pd.Timedelta(right)


@execute_node.register(
    (ops.TimestampAdd, ops.IntervalAdd), pd.Series, pd.Series
)
def execute_timestamp_interval_add_series_series(op, left, right, **kwargs):
    return left + right


@execute_node.register(ops.TimestampSub, timestamp_types, timedelta_types)
def execute_timestamp_sub_datetime_timedelta(op, left, right, **kwargs):
    return pd.Timestamp(left) - pd.Timedelta(right)


@execute_node.register(
    (ops.TimestampDiff, ops.TimestampSub), timestamp_types, pd.Series
)
def execute_timestamp_diff_sub_datetime_series(op, left, right, **kwargs):
    return pd.Timestamp(left) - right


@execute_node.register(ops.TimestampSub, pd.Series, timedelta_types)
def execute_timestamp_sub_series_timedelta(op, left, right, **kwargs):
    return left - pd.Timedelta(right)


@execute_node.register(
    (ops.TimestampDiff, ops.TimestampSub, ops.IntervalSubtract),
    pd.Series,
    pd.Series,
)
def execute_timestamp_diff_sub_series_series(op, left, right, **kwargs):
    return left - right


@execute_node.register(ops.TimestampDiff, timestamp_types, timestamp_types)
def execute_timestamp_diff_datetime_datetime(op, left, right, **kwargs):
    return pd.Timestamp(left) - pd.Timestamp(right)


@execute_node.register(ops.TimestampDiff, pd.Series, timestamp_types)
def execute_timestamp_diff_series_datetime(op, left, right, **kwargs):
    return left - pd.Timestamp(right)


@execute_node.register(
    ops.IntervalMultiply, pd.Series, numeric_types + (pd.Series,)
)
@execute_node.register(
    ops.IntervalFloorDivide,
    (pd.Timedelta, pd.Series),
    numeric_types + (pd.Series,),
)
def execute_interval_multiply_fdiv_series_numeric(op, left, right, **kwargs):
    return op.op(left, right)


@execute_node.register(ops.TimestampFromUNIX, (pd.Series,) + integer_types)
def execute_timestamp_from_unix(op, data, **kwargs):
    return pd.to_datetime(data, unit=op.unit)


@pre_execute.register(ops.TimestampNow)
@pre_execute.register(ops.TimestampNow, ibis.client.Client)
def pre_execute_timestamp_now(op, *args, **kwargs):
    return {op: pd.Timestamp('now')}


@execute_node.register(ops.DayOfWeekIndex, (str, datetime.date))
def execute_day_of_week_index_any(op, value, **kwargs):
    return pd.Timestamp(value).dayofweek


@execute_node.register(ops.DayOfWeekIndex, pd.Series)
def execute_day_of_week_index_series(op, data, **kwargs):
    return data.dt.dayofweek.astype(np.int16)


@execute_node.register(ops.DayOfWeekIndex, SeriesGroupBy)
def execute_day_of_week_index_series_group_by(op, data, **kwargs):
    groupings = data.grouper.groupings
    return data.obj.dt.dayofweek.astype(np.int16).groupby(groupings)


def day_name(obj):
    """Backwards compatible name of day getting function.

    Parameters
    ----------
    obj : Union[Series, pd.Timestamp]

    Returns
    -------
    str
        The name of the day corresponding to `obj`
    """
    try:
        return obj.day_name()
    except AttributeError:
        return obj.weekday_name


@execute_node.register(ops.DayOfWeekName, (str, datetime.date))
def execute_day_of_week_name_any(op, value, **kwargs):
    return day_name(pd.Timestamp(value))


@execute_node.register(ops.DayOfWeekName, pd.Series)
def execute_day_of_week_name_series(op, data, **kwargs):
    return day_name(data.dt)


@execute_node.register(ops.DayOfWeekName, SeriesGroupBy)
def execute_day_of_week_name_series_group_by(op, data, **kwargs):
    return day_name(data.obj.dt).groupby(data.grouper.groupings)


@execute_node.register(ops.DateSub, date_types, timedelta_types)
@execute_node.register((ops.DateDiff, ops.DateSub), date_types, pd.Series)
@execute_node.register(ops.DateSub, pd.Series, timedelta_types)
@execute_node.register((ops.DateDiff, ops.DateSub), pd.Series, pd.Series)
@execute_node.register(ops.DateDiff, date_types, date_types)
@execute_node.register(ops.DateDiff, pd.Series, date_types)
def execute_date_sub_diff(op, left, right, **kwargs):
    return left - right


@execute_node.register(ops.DateAdd, pd.Series, timedelta_types)
@execute_node.register(ops.DateAdd, timedelta_types, pd.Series)
@execute_node.register(ops.DateAdd, pd.Series, pd.Series)
@execute_node.register(ops.DateAdd, date_types, timedelta_types)
@execute_node.register(ops.DateAdd, timedelta_types, date_types)
@execute_node.register(ops.DateAdd, date_types, pd.Series)
@execute_node.register(ops.DateAdd, pd.Series, date_types)
def execute_date_add(op, left, right, **kwargs):
    return left + right



import pytest

import ibis

pytest.importorskip('clickhouse_driver')
pytestmark = pytest.mark.clickhouse


def test_column_ref_quoting(translate):
    schema = [('has a space', 'double')]
    table = ibis.table(schema)
    assert translate(table['has a space']) == '`has a space`'


def test_identifier_quoting(translate):
    schema = [('date', 'double'), ('table', 'string')]
    table = ibis.table(schema)
    assert translate(table['date']) == '`date`'
    assert translate(table['table']) == '`table`'


# TODO: fix it
# def test_named_expression(alltypes, translate):
#     a, b = alltypes.get_columns(['int_col', 'float_col'])
#     expr = ((a - b) * a).name('expr')

#     expected = '(`int_col` - `float_col`) * `int_col` AS `expr`'
#     assert translate(expr) == expected
